/*
 * dx2d.cpp
 *
 *  Created on: Jan 24, 2018
 *      Author: Hubert Vansteenhuyse
 *
 * This code normally belongs in winmain.c which is written in C, but,
 * Microsoft didn't fully implement C in the Direct2D and DirectWrite APIs.
 * MingW 64 implemented it as Microsoft did because written code would
 * otherwise only be compatible with MingW compilers and not with other compilers.
 *
 * This dll acts as an interface between winmain.c and Direct2D.
 *
 * Compile with a C++ compiler (e.g. MingW (64) gcc) with
 * libraries -ld2d1 -luuid -ldwrite
 */
#include <d2d1.h>
#include <dwrite.h>
#include <stdio.h>

#include "dx2d.h"
#include "platform.h"

#ifndef MAX
	#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif



static WCHAR __wc_buf[TXTOUT_MAX];


struct DX2D_GLOBALS {
	iface_plat_dll			iface;

	float					dpix, dpiy;

	HRESULT 				dxResult;
	ID2D1Factory			*dxFactory;
	D2D1_FACTORY_OPTIONS 	dxFactoryOptions;

	ID2D1HwndRenderTarget					*dxRenderTarget;
	D2D1_RENDER_TARGET_PROPERTIES 			dxRenderProps;
	D2D1_HWND_RENDER_TARGET_PROPERTIES		dxHwndRenderProps;

	ID2D1SolidColorBrush					*Brush, *Pen;
	D2D1_BRUSH_PROPERTIES					dxBrushProps, dxPenProps;
	D2D1_COLOR_F 							colBrush, colPen;

	D2D1_STROKE_STYLE_PROPERTIES 			dxStrokeProps;
	ID2D1StrokeStyle						*dxStrokeStyle;


	IDWriteFactory							*dwFactory;
	IDWriteRenderingParams					*dwParams;
	IDWriteTextFormat						*dwFormat;

	char									cFontName[TXTOUT_MAX];
	WCHAR 									wcFontName[TXTOUT_MAX];
	int										FontSize, FontW, FontH;
	int										swBold, swItalic, swOpaque;

	//I know there is one but I can't find where the static Identity function is declared (D2D1Identity ?)
	D2D1_MATRIX_3X2_F Identity3X2(){
		D2D1_MATRIX_3X2_F i;
		i._12 = i._21 = i._31 = i._32 = 0;
		i._11 = i._22 = 1;
		return i;
	}

	void ReleaseFactories(void){
		ReleaseStrokeStyle();
		if(dxFactory) dxFactory->Release();
		dxFactory = 0;
		if(dwFactory) dwFactory->Release();
		dwFactory = 0;
	}

	void ReleaseBrush(void){
		if(Brush) Brush->Release();
		Brush = 0;
	}

	void ReleasePen(void){
		if(Pen) Pen->Release();
		Pen = 0;
	}

	void ReleaseStrokeStyle(void){
		if(dxStrokeStyle) dxStrokeStyle->Release();
		dxStrokeStyle = 0;
	}


	void ReleaseRenderTarget(void){
		ReleasePen();
		ReleaseBrush();

		if(dxRenderTarget) dxRenderTarget->Release();
		dxRenderTarget = 0;
	}

	void ReleaseTextFormat(){
		if(dwFormat) dwFormat->Release();
		dwFormat = 0;
	}

	/*********
	 * every x, y, rectangle, radius, ... must be converted from windows parameters to DIP
	 * when using direct 2D rendering methods
	 *********/
	//https://docs.microsoft.com/en-us/windows/desktop/learnwin32/dpi-and-device-independent-pixels
	inline float dip_x(float x) { return x / dpix; }
	inline float dip_y(float y) { return y / dpiy; }

	inline float dip(float n) {
		float 	x = dip_x(n),
				y = dip_y(n);
		return MIN(x, y);
	}

	inline void dip_rect(D2D1_RECT_F &r){
		r.left = dip_x(r.left);
		r.right = dip_x(r.right);
		r.top = dip_y(r.top);
		r.bottom = dip_y(r.bottom);
	}
};

DX2D_GLOBALS globals;

HRESULT WINAPI	dx2d_init(iface_plat_dll *iface, int swDoubleBuffer){
	memset(&globals, 0, sizeof(DX2D_GLOBALS));

	globals.iface = *iface;

	//Direct2D
	globals.dxFactoryOptions.debugLevel = D2D1_DEBUG_LEVEL_NONE;
	globals.dxResult = D2D1CreateFactory(
			D2D1_FACTORY_TYPE_SINGLE_THREADED,
			__uuidof(ID2D1Factory),
			&globals.dxFactoryOptions,
			(void**)(&globals.dxFactory));

	if(!SUCCEEDED(globals.dxResult) || !globals.dxFactory){
		return globals.dxResult;
	}

	globals.dxStrokeProps.startCap = D2D1_CAP_STYLE_FLAT;
	globals.dxStrokeProps.endCap = D2D1_CAP_STYLE_FLAT;
	globals.dxStrokeProps.dashOffset = -1.f;
	globals.dxStrokeProps.dashStyle = D2D1_DASH_STYLE_DASH;
	globals.dxStrokeProps.lineJoin = D2D1_LINE_JOIN_MITER;
	globals.dxStrokeProps.miterLimit = 1.f;
	globals.dxFactory->CreateStrokeStyle(&globals.dxStrokeProps, NULL, 0, &globals.dxStrokeStyle);

	//DirectWrite
	globals.dxResult = DWriteCreateFactory(
				DWRITE_FACTORY_TYPE_SHARED,
				__uuidof(IDWriteFactory),
				(IUnknown**)(&globals.dwFactory));

	if(!SUCCEEDED(globals.dxResult) || !globals.dwFactory){
		globals.ReleaseFactories();
		return globals.dxResult;
	}

	//DirectWrite default parameters
	globals.dxResult = globals.dwFactory->CreateRenderingParams(&globals.dwParams);
	if(!SUCCEEDED(globals.dxResult) || !globals.dwFactory){
		globals.ReleaseFactories();
		return globals.dxResult;
	}

	//the directx props
	globals.dxFactory->GetDesktopDpi(&globals.dxRenderProps.dpiX, &globals.dxRenderProps.dpiY);
	globals.dpix = globals.dxRenderProps.dpiX / 96.0;
	globals.dpiy = globals.dxRenderProps.dpiY / 96.0;

	globals.dxRenderProps.type = D2D1_RENDER_TARGET_TYPE_DEFAULT;
	globals.dxRenderProps.pixelFormat.format = DXGI_FORMAT_UNKNOWN;
	globals.dxRenderProps.pixelFormat.alphaMode = D2D1_ALPHA_MODE_UNKNOWN;
	globals.dxRenderProps.usage = D2D1_RENDER_TARGET_USAGE_NONE;
	globals.dxRenderProps.minLevel = D2D1_FEATURE_LEVEL_DEFAULT;

	//the hwnd directx props
	globals.dxHwndRenderProps.hwnd = 0;
	globals.dxHwndRenderProps.pixelSize.width = globals.dxHwndRenderProps.pixelSize.height = 0;
	globals.dxHwndRenderProps.presentOptions = swDoubleBuffer ? D2D1_PRESENT_OPTIONS_RETAIN_CONTENTS : D2D1_PRESENT_OPTIONS_IMMEDIATELY;

	//Pen and Brush props
	//(hoping that there are no other hidden matrices in these DirectX classes)
	globals.dxPenProps.transform = globals.Identity3X2();
	globals.dxBrushProps.transform = globals.Identity3X2();
	globals.dxPenProps.opacity = 1.f;
	globals.dxBrushProps.opacity = 1.f;

	return globals.dxResult;
}

HRESULT WINAPI	dx2d_exit(void){
	globals.ReleaseRenderTarget();
	globals.ReleaseFactories();
	return S_OK;
}

HRESULT WINAPI 		dx2d_render_start	(UINT32 w, UINT32 h, HWND hwnd){
	D2D1_SIZE_U size;

	if( !globals.dxRenderTarget || globals.dxHwndRenderProps.hwnd != hwnd || globals.dxHwndRenderProps.pixelSize.width != w || globals.dxHwndRenderProps.pixelSize.height != h ){
		//force the creation of a new renderer
		if(globals.dxHwndRenderProps.hwnd != hwnd)
			globals.ReleaseRenderTarget();


		globals.dxHwndRenderProps.pixelSize.width = w;
		globals.dxHwndRenderProps.pixelSize.height = h;
		if(!globals.dxRenderTarget){
			//(re) create the renderer
			globals.dxHwndRenderProps.hwnd = hwnd;
			globals.dxResult = globals.dxFactory->CreateHwndRenderTarget(&globals.dxRenderProps, &globals.dxHwndRenderProps, &globals.dxRenderTarget);
			if(!SUCCEEDED(globals.dxResult)){
				globals.ReleaseRenderTarget();
				return globals.dxResult;
			}
		}

		//resize only
		size = D2D1::SizeU(w, h);
		globals.dxResult = globals.dxRenderTarget->Resize(size);
		if(!SUCCEEDED(globals.dxResult)){
			globals.ReleaseRenderTarget();
			return globals.dxResult;
		}

		globals.dxRenderTarget->SetAntialiasMode(D2D1_ANTIALIAS_MODE_PER_PRIMITIVE);
		globals.dxRenderTarget->SetTextAntialiasMode(D2D1_TEXT_ANTIALIAS_MODE_CLEARTYPE);

	}

	if(globals.dxRenderTarget){
		globals.dxRenderTarget->BeginDraw();
		return S_OK;
	}

	return S_FALSE;
}

HRESULT WINAPI 		dx2d_render_end		(void){
	if(globals.dxRenderTarget){
		globals.dxRenderTarget->EndDraw(0, 0);
		return S_OK;
	}
	return S_FALSE;
}

HRESULT WINAPI 		dx2d_color_pen		(float col_maxval, float r, float g, float b, float a){
	if(globals.dxRenderTarget){
		globals.colPen.r = r / col_maxval;
		globals.colPen.g = g / col_maxval;
		globals.colPen.b = b / col_maxval;
		globals.colPen.a = a;
		if(globals.Pen){
			globals.Pen->SetColor(globals.colPen);
		} else {
			globals.dxResult = globals.dxRenderTarget->CreateSolidColorBrush(globals.colPen, globals.dxPenProps, &globals.Pen);
			return globals.dxResult;
		}
		return S_OK;
	}
	return S_FALSE;
}

HRESULT WINAPI 		dx2d_color_brush	(float col_maxval, float r, float g, float b, float a){
	if(globals.dxRenderTarget){
		globals.colBrush.r = r / col_maxval;
		globals.colBrush.g = g / col_maxval;
		globals.colBrush.b = b / col_maxval;
		globals.colBrush.a = a;
		if(globals.Brush){
			globals.Brush->SetColor(globals.colBrush);
		} else {
			globals.dxResult = globals.dxRenderTarget->CreateSolidColorBrush(globals.colBrush, globals.dxBrushProps, &globals.Brush);
			return globals.dxResult;
		}
		return S_OK;
	}
	return S_FALSE;
}

int		WINAPI		dx2d_fontw			(void){
	return globals.FontW;
}
int		WINAPI		dx2d_fonth			(void){
	return globals.FontH;
}

HRESULT	WINAPI		dx2d_usefont		(const char *name, int size, int swBold, int swItalic){
	globals.dxResult = S_FALSE;
	if(globals.dwFactory && name){
		globals.iface.WC((char *)name, globals.wcFontName);

		globals.FontSize = size;

		float fontsze = globals.dip(size);

		globals.ReleaseTextFormat();
		globals.dwFactory->CreateTextFormat(
				globals.wcFontName,
				NULL,
				swBold ? DWRITE_FONT_WEIGHT_BOLD : DWRITE_FONT_WEIGHT_NORMAL,
				swItalic ? DWRITE_FONT_STYLE_ITALIC : DWRITE_FONT_STYLE_NORMAL,
				DWRITE_FONT_STRETCH_NORMAL,
				fontsze,
				L"",
				&globals.dwFormat);
		if(globals.dwFormat){
			globals.dxResult = S_OK;
			//globals.FontSize = globals.dwFormat->GetFontSize() + 0.5;
			globals.dwFormat->SetWordWrapping(DWRITE_WORD_WRAPPING_NO_WRAP);
		}
		//measure char W
		/*
		int w = 0, h = 0;
		dx2d_lentextw(globals.iface.WC((char*)"W", 0), 1, &w, &h);
		globals.FontW = MAX(w, globals.FontSize);
		globals.FontH = MAX(h, globals.FontSize);
*/
		dx2d_lentextw(globals.iface.WC((char*)"W", 0), 1, &globals.FontW, &globals.FontH);

		globals.swBold = swBold;
		globals.swItalic = swItalic;
	}
	return globals.dxResult;
}

void	WINAPI		dx2d_set_text		(int swBold, int swItalic, int swOpaque){
	char tmp[TXTOUT_MAX];
	if(globals.swBold != swBold || globals.swItalic != swItalic){
		snprintf(tmp, TXTOUT_MAX - 1, "%s", globals.cFontName);
		dx2d_usefont(tmp, globals.FontSize, swBold, swItalic);
	}
	globals.swOpaque = swOpaque;
}

void	WINAPI		dx2d_lentextw		(WCHAR *str, int n, int *w, int *h){
	IDWriteTextLayout 	*layout = 0;
	DWRITE_TEXT_METRICS	tm;
	if(w) *w = 0;
	if(h) *h = 0;
	if(n && globals.dwFactory && globals.dwFormat){
		globals.dwFactory->CreateTextLayout(
				str,
				n,
				globals.dwFormat,
				globals.dxHwndRenderProps.pixelSize.width,
				globals.dxHwndRenderProps.pixelSize.height,
				&layout);
		if(layout){
			if(SUCCEEDED(layout->GetMetrics(&tm))){
				if(w) *w = tm.widthIncludingTrailingWhitespace + globals.FontW * 0.5;
				if(h) *h = tm.height * 1.25;
			}
			layout->Release();
		}
	}
}

void	WINAPI		dx2d_outtextw		(int x, int y, WCHAR *str, int n){
	if(n && globals.dxRenderTarget && globals.dwFormat && globals.Pen && globals.Brush){
		D2D1_RECT_F r;
		int w, h;
		dx2d_lentextw(str, n, &w, &h);
		r.left = x;
		r.top = y;
		r.right = x + w;
		r.bottom = y + h;

		globals.dip_rect(r);

		if(globals.swOpaque){
			globals.dxRenderTarget->FillRectangle(r, globals.Brush);
		}

		globals.dxRenderTarget->DrawTextA(
				str,
				n,
				globals.dwFormat,
				r,
				globals.Pen,
				D2D1_DRAW_TEXT_OPTIONS_NONE,
				DWRITE_MEASURING_MODE_NATURAL);
	}
}

void	WINAPI		dx2d_cls			(void){
	if(globals.dxRenderTarget){
		globals.dxRenderTarget->Clear(globals.colBrush);
	}
}

void 	WINAPI		dx2d_line			(int x1, int y1, int x2, int y2){
	if(globals.dxRenderTarget && globals.Pen){
		globals.dxRenderTarget->DrawLine(
				D2D1::Point2F((FLOAT)globals.dip_x(x1), (FLOAT)globals.dip_y(y1)),
				D2D1::Point2F((FLOAT)globals.dip_x(x2), (FLOAT)globals.dip_y(y2)),
				globals.Pen,
				1.f);
	}
}

void 	WINAPI		dx2d_dashline		(int x1, int y1, int x2, int y2){
	if(globals.dxRenderTarget && globals.Pen){
		globals.dxRenderTarget->DrawLine(
				D2D1::Point2F((FLOAT)globals.dip_x(x1), (FLOAT)globals.dip_y(y1)),
				D2D1::Point2F((FLOAT)globals.dip_x(x2), (FLOAT)globals.dip_y(y2)),
				globals.Pen,
				1.f,
				globals.dxStrokeStyle);
	}
}


void 	WINAPI		dx2d_rectangle		(int swFilled, int x, int y, int wdt, int hgt){
	if(globals.dxRenderTarget && globals.dwFormat && globals.Pen && globals.Brush){
		D2D1_RECT_F r;
		r.left = x;
		r.top = y;
		r.right = x + wdt;
		r.bottom = y + hgt;

		globals.dip_rect(r);

		if(swFilled){
			globals.dxRenderTarget->FillRectangle(r, globals.Brush);
		} else {
			globals.dxRenderTarget->DrawRectangle(r, globals.Pen);
		}
	}
}

void 	WINAPI		dx2d_circle			(int swFilled, int x, int y, int r){
	if(globals.dxRenderTarget && globals.dwFormat && globals.Pen && globals.Brush){
		D2D1_ELLIPSE	e;
		e.point.x = globals.dip_x(x);
		e.point.y = globals.dip_y(y);
		e.radiusX = e.radiusY = globals.dip(r);

		if(swFilled){
			globals.dxRenderTarget->FillEllipse(e, globals.Brush);
		} else {
			globals.dxRenderTarget->DrawEllipse(e, globals.Pen);
		}

	}
}

void 	WINAPI		dx2d_polygon		(int swFilled, int elements, double *x, double *y){
	if(elements && x && y && globals.dxFactory && globals.dxRenderTarget){
		int n = 0;
		ID2D1PathGeometry *path = 0;
		ID2D1GeometrySink *sink = 0;
		D2D1_POINT_2F p;

		globals.dxFactory->CreatePathGeometry(&path);
		if(!path) return;

		path->Open(&sink);
		if(sink){
			sink->BeginFigure(D2D1::Point2F((FLOAT)globals.dip_x(x[n]), (FLOAT)globals.dip_y(y[n])), swFilled ? D2D1_FIGURE_BEGIN_FILLED : D2D1_FIGURE_BEGIN_HOLLOW);
			n++;
			while(n < elements){
				sink->AddLine(D2D1::Point2F((FLOAT)globals.dip_x(x[n]), (FLOAT)globals.dip_y(y[n])));
				n++;
			}

			sink->EndFigure(D2D1_FIGURE_END_CLOSED);
			sink->Close();

			if(swFilled){
				globals.dxRenderTarget->FillGeometry(path, globals.Brush);
			} else {
				globals.dxRenderTarget->DrawGeometry(path, globals.Pen, 1.f);
			}

			sink->Release();
		}

		path->Release();


	}
}

void 	WINAPI		dx2d_linepath		(int elements, double *x, double *y){
	if(globals.dxRenderTarget && globals.Pen && elements > 1){
		for(int i = 0; i < elements - 1; i++){
			globals.dxRenderTarget->DrawLine(
					D2D1::Point2F((FLOAT)globals.dip_x(x[i]), (FLOAT)globals.dip_y(y[i])),
					D2D1::Point2F((FLOAT)globals.dip_x(x[i + 1]), (FLOAT)globals.dip_y(y[i + 1])),
					globals.Pen,
					1.f);
		}
	}

}

