/*
    Copyright (C) 2000, 2020 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

/**
 * Implementation for Windows GDI 32.
 * Compatible with any Windows version fast but with simple RGB graphics
 * LINKER: nothing
 * DEF: PLATFORM_GDI or compiler -DPLATFORM_GDI
 * -----------------------------------
 * Implementation for Windows GDI PLUS FLAT (for c language)
 * Compatible with most Windows version slow but with better RGBA graphics
 * HEADER: gdiplus.h
 * LINKER: -lgdiplus
 * DEF: PLATFORM_GDI_PLUS or compiler -DPLATORM_GDI_PLUS
 * -----------------------------------
 * Implementation for Windows DirectX 2D
 * Compatible on more recent Windows versions fast gaming graphics
 * HEADER: d2d1.h
 * LINKER: -ld2d1 -luuid
 * DEF: PLATFORM_DIRECT2D or compiler -DPLATORM_DIRECTX
 */

//#define PLATFORM_GDI
//#define PLATFORM_GDI_PLUS
//#define PLATFORM_DIRECT2D

#define _WIN32_IE    0x0500
#define _WIN32_WINNT 0x0501
#define TABCLOSE_TEXT "x"

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <commdlg.h>
#include <xpsprint.h>


#include <stdio.h>
#include <stdlib.h>
#include "platform.h"

#include <sys/stat.h>
#include <sys/types.h>

#include "wchar.h"

#ifdef PLATFORM_GDI_PLUS
	#include <gdiplus.h>

	int __font_hgt = 0, __font_wdt = 0;

	#define MAKE_ARGB(r, g, b , a)	(ARGB) ((((DWORD) (a)) << 24) | (((DWORD) (r)) << 16) | (((DWORD) (g)) << 8) | ((DWORD) (b)))

	GpGraphics					*gdpGraphics = 0;
	ULONG_PTR					gdpToken = 0;
	struct GdiplusStartupInput	gdpInput;
#endif

#ifdef PLATFORM_DIRECT2D
	#include "dx2d.h"
	iface_plat_dll	dllDX2D_iface;

	HRESULT 		dxResult = 0;
	HMODULE 		hDx2d = 0;

	HRESULT 		WINAPI 	(CALLBACK* dllDX2D_init)(iface_plat_dll *iface, int swDoubleBuffer) = 0;
	HRESULT 		WINAPI	(CALLBACK* dllDX2D_exit)(PLATDX2D) = 0;

	HRESULT 		WINAPI 	(CALLBACK* dllDX2D_render_start) (UINT32 w, UINT32 h, HWND hwnd) = 0;
	HRESULT 		WINAPI 	(CALLBACK* dllDX2D_render_end)(void) = 0;

	HRESULT 		WINAPI 	(CALLBACK* dllDX2D_color_pen)(float col_maxval, float r, float g, float b, float a) = 0;
	HRESULT 		WINAPI 	(CALLBACK* dllDX2D_color_brush)(float col_maxval, float r, float g, float b, float a) = 0;

	int				WINAPI	(CALLBACK* dllDX2D_fontw)(void) = 0;
	int				WINAPI	(CALLBACK* dllDX2D_fonth)(void) = 0;
	HRESULT			WINAPI	(CALLBACK* dllDX2D_usefont)(const char *name, int size, int swBold, int swItalic) = 0;
	void			WINAPI	(CALLBACK* dllDX2D_set_text)(int swBold, int swItalic, int swOpaque) = 0;
	void			WINAPI	(CALLBACK* dllDX2D_lentextw)(WCHAR *str, int n, int *w, int *h) = 0;
	void			WINAPI	(CALLBACK* dllDX2D_outtextw)(int x, int y, WCHAR *str, int n) = 0;

	void			WINAPI	(CALLBACK* dllDX2D_cls)(void) = 0;
	void 			WINAPI	(CALLBACK* dllDX2D_line)(int x1, int y1, int x2, int y2) = 0;
	void 			WINAPI	(CALLBACK* dllDX2D_dashline)(int x1, int y1, int x2, int y2) = 0;

	void 			WINAPI	(CALLBACK* dllDX2D_rectangle)(int swFilled, int x, int y, int wdt, int hgt) = 0;
	void 			WINAPI	(CALLBACK* dllDX2D_circle)(int swFilled, int x, int y, int r) = 0;
	void 			WINAPI	(CALLBACK* dllDX2D_polygon)(int swFilled, int elements, double *x, double *y) = 0;
	void 			WINAPI	(CALLBACK* dllDX2D_linepath)(int elements, double *x, double *y) = 0;

#endif

#define WC_CUSTOMDIALOG "DialogWindowClass"


typedef struct _local_MEMORYSTATUSEX {
  DWORD     dwLength;
  DWORD     dwMemoryLoad;
  DWORDLONG ullTotalPhys;
  DWORDLONG ullAvailPhys;
  DWORDLONG ullTotalPageFile;
  DWORDLONG ullAvailPageFile;
  DWORDLONG ullTotalVirtual;
  DWORDLONG ullAvailVirtual;
  DWORDLONG ullAvailExtendedVirtual;
} _local_MEMORYSTATUSEX, *_local_LPMEMORYSTATUSEX;

typedef struct THREADDATA {
	__timer_routine	tmfunc;	/*timer routine to be called*/
	void			*tdata;
	struct timeval 	tv;
	float			msec;
}thread_data_t;
volatile LIST* lstTimer = 0;

HCURSOR cursors[3];
float __alpha = 1, __intensity = 1;
int		col_maxval = 255;
int __def_font_sze = 12;

void doevents();
int maxx(void);
int maxy(void);
void resize_children();
void on_toolbar_item_clicked(CTL_TOOLBAR_ITEM *tbitem);
void __create_caret(int swOn);

/*gdi32.dll routines*/
HMODULE hWinGdi = 0;
BOOL WINAPI (CALLBACK* dllCreateFont)(int nHeight, int nWidth, int nEscapement, int nOrientation, int fnWeight, DWORD fdwItalic, DWORD fdwUnderline, DWORD fdwStrikeOut, DWORD fdwCharSet, DWORD fdwOutputPrecision, DWORD fdwClipPrecision, DWORD fdwQuality, DWORD fdwPitchAndFamily, LPCTSTR lpszFace) = 0;
BOOL WINAPI (CALLBACK* dllGetTextFace)(HDC hdc, int n, LPTSTR name) = 0;
BOOL WINAPI (CALLBACK* dllLineTo)(HDC hdc, int nXEnd, int nYEnd) = 0;
BOOL WINAPI (CALLBACK* dllMoveToEx)(HDC hdc, int X, int Y, LPPOINT lpPoint) = 0;
BOOL WINAPI (CALLBACK* dllRectangle)(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect) = 0;
BOOL WINAPI (CALLBACK* dllPolygon)(HDC hdc, const POINT *lpPoints, int nCount) = 0;
BOOL WINAPI (CALLBACK* dllPolyline)(HDC hdc, const POINT *lpPoints, int nCount) = 0;
BOOL WINAPI (CALLBACK* dllEllipse)(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect) = 0;
BOOL WINAPI (CALLBACK* dllArc)(HDC hdc, int nLeftRect, int nTopRect, int nRightRect, int nBottomRect, int nXStartArc, int nYStartArc, int nXEndArc, int nYEndArc) = 0;
BOOL WINAPI (CALLBACK* dllAngleArc)(HDC hdc, int x, int y, DWORD r, FLOAT StartAngle, FLOAT SweepAngle) = 0;


BOOL WINAPI (CALLBACK* dllTextOut)(HDC hdc, int x, int y, char*, int len) = 0;
BOOL WINAPI (CALLBACK* dllTextOutW)(HDC hdc, int x, int y, wchar_t*, int len) = 0;
BOOL WINAPI (CALLBACK* dllGetTextExtentPoint)(HDC hdc, char *str, int c, LPSIZE lpSize);
BOOL WINAPI (CALLBACK* dllGetTextExtentPointW)(HDC hdc, wchar_t* lpString, int c, LPSIZE lpSize);
BOOL WINAPI (CALLBACK* dllSetPixel)(HDC hdc, int X, int Y, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreatePen)(int fnPenStyle, int nWidth, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreateSolidBrush)(COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllSelectObject)(HDC hdc, HGDIOBJ hgdiobj) = 0;
BOOL WINAPI (CALLBACK* dllDeleteObject)(HGDIOBJ hObject) = 0;
BOOL WINAPI (CALLBACK* dllSetTextColor)(HDC hdc, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllSetBkColor)(HDC hdc, COLORREF crColor) = 0;
BOOL WINAPI (CALLBACK* dllCreateCompatibleDC)(HDC hdc) = 0;
BOOL WINAPI (CALLBACK* dllBitBlt)(HDC hdcDest, int nXDest, int nYDest, int nWidth, int nHeight,	HDC hdcSrc, int nXSrc, int nYSrc, DWORD dwRop) = 0;
BOOL WINAPI (CALLBACK* dllCreateCompatibleBitmap)(HDC hdc, int nWidth, int nHeight) = 0;
BOOL WINAPI (CALLBACK* dllGetTextMetrics)(HDC hdc, LPTEXTMETRIC lptm) = 0;
BOOL WINAPI (CALLBACK* dllSetBkMode)(HDC hdc, int mode) = 0;


/*user32.dll routines*/
HMODULE hUser32 = 0;
BOOL WINAPI (CALLBACK* dllFillRect)(HDC hdc, RECT *rect, HBRUSH hbr) = 0;

/*Comctl32.dll routines*/
HMODULE hWinCtrl = 0;
INITCOMMONCONTROLSEX initCtrl;
BOOL WINAPI (CALLBACK* dllInitCommonControlsEx)(const LPINITCOMMONCONTROLSEX lpInitCtrls) = 0;
BOOL WINAPI (CALLBACK* dllImageList_Create)(int cx, int cy, UINT flags, int cInitial, int cGrow) = 0;
BOOL WINAPI (CALLBACK* dllImageList_Destroy)(HIMAGELIST himl) = 0;

/*Comdlg32.dll routines*/
HMODULE hWinDlg = 0;
BOOL WINAPI (CALLBACK* dllGetOpenFileName)(LPOPENFILENAME lpofn) = 0;
BOOL WINAPI (CALLBACK* dllGetSaveFileName)(LPOPENFILENAME lpofn) = 0;
BOOL WINAPI (CALLBACK* dllPrintDlg)(LPPRINTDLG lppd) = 0;
BOOL WINAPI (CALLBACK* dllChooseFont)(LPCHOOSEFONT lpcf) = 0;

/*kernel32.dll routines*/
HMODULE hKernel32 = 0;
BOOL WINAPI (CALLBACK* dllGlobalMemoryStatusEx)(_local_LPMEMORYSTATUSEX lpBuffer) = 0;

typedef struct WINDC{
	HDC		windc, bmpdc, oldbmpdc;
	HBITMAP	bmp;
}WINDC;

WINDC mdc;

int kbd = 0, mouseX = 0, mouseY = 0;
IPLAT mouseB = IPLAT_NULL;

#ifdef PLATFORM_GDI_PLUS
	GpPen			*Pen = 0;
	GpSolidFill 	*Brush = 0;
	GpFont			*Font = 0;
	GpStringFormat 	*gdpSF = 0;
#endif

#ifdef PLATFORM_GDI
	HGDIOBJ OldPen = 0, OldBrush = 0, OldFont = 0;
	HPEN Pen = 0;
	HBRUSH Brush = 0;
	HFONT Font = 0;
	COLORREF fcol, bcol;
#endif

int 	__got_focus = 0,
		__sw_showCaret = 0,
		__caret_created = 0,
		__sw_caret_moved = 0,
		__caret_hgt = 0,
		__caret_wdt = 0,
		__caret_gap = 0;

IPLAT 	__caret_mode = CARET_NORMAL;

int 	__sw_opaque = 0,
		__sw_reverse = 0,
		__sw_bold = 0,
		__sw_italic = 0,
		__sw_underlined = 0;

UINT sbarflags = 0;
int swWinExit = 0, swUpdateColors = 1;

HINSTANCE 	hinstMain = 0;
HWND 		hwndMain = 0,
			hwndChild = 0,
			hwndTab = 0,
			hwndTabCloseBtn = 0,
			hwndToolBar = 0,
			hwndStatus = 0;

WNDCLASSEXW 	wndclass,
				childclass,
				tabclassbutton,
				wndClassDialog;

#define CLASSNAME_MAX 128
WCHAR 		szMainWndClass[CLASSNAME_MAX],			//main window class name
			szToolbarWndClass[CLASSNAME_MAX],		//toolbar
			szTabControlWndClass[CLASSNAME_MAX],	//tab window
			szChildWndClass[CLASSNAME_MAX],			//tab child windows
			szTabButtonClass[CLASSNAME_MAX],		//tab child close button
			szTabStatusBarClass[CLASSNAME_MAX],		//status bar
			szDialogWindowClass[CLASSNAME_MAX];



HIMAGELIST	hilstToolBar = 0;	//image list for the toolbar

UINT_PTR	__timer_handler = 0;
volatile int __is_drawing = 0;

HFONT	hTabFont = 0, hDialogFont = 0;
int		TabHeight = 24, DialogSepHor = 0, DialogSepVer = 0;
NONCLIENTMETRICS	ncm;
PLATFORM_APP *fwplat = 0;

LIST 	*lstToolBarItems = 0,
		*lstTabPages = 0,
		*lstStatusBar = 0;

/**
 * System colors
 **/
C_RGBA		color_fg, color_bg,				/*system color client text area*/
			color_btn_bg, color_btn_fg,		/*system color button*/
			color_sel_bg, color_sel_fg,		/*system color selected text*/
			color_tip_bg, color_tip_fg,		/*system color tool tip*/

			colorf, colorb;					/*currently selected colors*/

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** GENERAL FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/

/*
 * If an UTF16 string is passed then the result is unpredictable.
 * Microsoft uses UTF16 but is incapable of telling whether a string is UTF16 or not.
 *
 * Everything that is not UTF8 is converted into local code page UTF16.
 * Everything that is UTF8 is converted into UTF16.
 */
WCHAR __impl_wc[TXTOUT_MAX];
WCHAR* WC(char *txt, WCHAR *out){
	int n = txt ? strlen(txt) : 0;
	int swUTF8;
	if(out == 0) out = &__impl_wc;
	if(n){
		swUTF8 = plat_utf8_valid(txt);
		n = MultiByteToWideChar(swUTF8 ? CP_UTF8 : CP_THREAD_ACP, 0, txt, n, out, TXTOUT_MAX);
	}
	out[n] = (WCHAR)0;

	return out;
}

/*
 * Converts UTF16 back to UTF8
 */
char __impl_uni[TXTOUT_MAX];
char *UTF8(WCHAR *txt, char *out){
	int n = txt ? wcslen(txt) : 0;
	if(out == 0) out = &__impl_uni;
	if(n){
		n = WideCharToMultiByte(CP_UTF8, 0, txt, n, out, TXTOUT_MAX, 0, 0);
	}
	out[n] = (char)0;

	return out;
}

int __win_redraw(PLAT_RECT *r){
	RECT wr;
	if(hwndChild){
		if(r){
			wr.left = r->x; wr.right = r->xx;
			wr.top = r->y; wr.bottom = r->yy;
			return (int)RedrawWindow(hwndChild, &wr, 0, RDW_INVALIDATE);
		}
		return (int)RedrawWindow(hwndChild, 0, 0, RDW_INVALIDATE);
	}
	return 0;
}

void __begin_draw(HDC hdc){
	DWORD ncolor;

	if(!__is_drawing && !fwplat->gui.sysdata){
		__is_drawing = 1;

		#ifndef PLATFORM_DIRECT2D		//valid for all except for direct x
			mdc.windc = hdc;
			if(fwplat->sw_double_buffer){
				mdc.bmpdc = (HDC)dllCreateCompatibleDC(mdc.windc);
				mdc.bmp = (HBITMAP)dllCreateCompatibleBitmap(mdc.windc, maxx(), maxy());
			} else {
				mdc.bmpdc = mdc.windc;
			}
			mdc.oldbmpdc = (HDC)dllSelectObject(mdc.bmpdc, mdc.bmp);
		#endif

		fwplat->gui.sysdata = &mdc;

		#ifdef PLATFORM_GDI_PLUS
			GdipCreateFromHDC(mdc.bmpdc, &gdpGraphics);
			GdipSetCompositingQuality(gdpGraphics, CompositingQualityHighQuality);
			GdipSetSmoothingMode(gdpGraphics, SmoothingModeHighQuality);
			GdipSetTextRenderingHint(gdpGraphics, TextRenderingHintSystemDefault);
		#endif

		#ifdef PLATFORM_DIRECT2D
		dxResult = dllDX2D_render_start(maxx(), maxy(), hwndChild);
		#endif

		if(__sw_showCaret && __got_focus) HideCaret(hwndChild);

		if(swUpdateColors){
			fwplat->gui.color_maxval = col_maxval;

			ncolor = GetSysColor(COLOR_WINDOW);
			color_bg.r = GetRValue(ncolor);
			color_bg.g = GetGValue(ncolor);
			color_bg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_WINDOWTEXT);
			color_fg.r = GetRValue(ncolor);
			color_fg.g = GetGValue(ncolor);
			color_fg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_HIGHLIGHT);
			color_sel_bg.r = GetRValue(ncolor);
			color_sel_bg.g = GetGValue(ncolor);
			color_sel_bg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_HIGHLIGHTTEXT);
			color_sel_fg.r = GetRValue(ncolor);
			color_sel_fg.g = GetGValue(ncolor);
			color_sel_fg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_INFOBK);/*24*/
			color_tip_bg.r = GetRValue(ncolor);
			color_tip_bg.g = GetGValue(ncolor);
			color_tip_bg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_INFOTEXT);/*23*/
			color_tip_fg.r = GetRValue(ncolor);
			color_tip_fg.g = GetGValue(ncolor);
			color_tip_fg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_BTNFACE);/*15*/
			color_btn_bg.r = GetRValue(ncolor);
			color_btn_bg.g = GetGValue(ncolor);
			color_btn_bg.b = GetBValue(ncolor);

			ncolor = GetSysColor(COLOR_BTNTEXT);
			color_btn_fg.r = GetRValue(ncolor);
			color_btn_fg.g = GetGValue(ncolor);
			color_btn_fg.b = GetBValue(ncolor);

			swUpdateColors = 0;
		}
	}
}
void __end_draw(PAINTSTRUCT *ps){
	if(__is_drawing && fwplat->gui.sysdata){
		#ifdef PLATFORM_DIRECT2D
		dllDX2D_render_end();
		#endif

		if(mdc.windc){
			#ifdef PLATFORM_GDI_PLUS
				if(gdpGraphics) GdipDeleteGraphics(gdpGraphics);
				gdpGraphics = 0;
			#endif

			#ifndef PLATFORM_DIRECT2D		//valid for all except for direct x
				if(fwplat->sw_double_buffer) dllBitBlt(mdc.windc, ps->rcPaint.left, ps->rcPaint.top, ps->rcPaint.right - ps->rcPaint.left, ps->rcPaint.bottom - ps->rcPaint.top, mdc.bmpdc, ps->rcPaint.left, ps->rcPaint.top, SRCCOPY);
			#endif

			if(__sw_showCaret && __got_focus) ShowCaret(hwndChild);

			#ifdef PLATFORM_GDI
				if(OldPen) dllSelectObject(mdc.bmpdc, OldPen);
				OldPen = 0;
				if(Pen) dllDeleteObject(Pen);
				Pen = 0;

				if(OldBrush) dllSelectObject(mdc.bmpdc, OldBrush);
				OldBrush = 0;
				if(Brush) dllDeleteObject(Brush);
				Brush = 0;

				if(OldFont) dllSelectObject(mdc.bmpdc, OldFont);
				OldFont = 0;
				if(Font) dllDeleteObject(Font);
				Font = 0;
			#endif

			#ifndef PLATFORM_DIRECT2D		//valid for all except for direct x
				if(fwplat->sw_double_buffer){
					dllDeleteObject(mdc.oldbmpdc);
					dllDeleteObject(mdc.bmp);
				}
				dllDeleteObject(mdc.bmpdc);
			#endif

			mdc.bmpdc = mdc.windc = 0;
		}

		fwplat->gui.sysdata = 0;
		fwplat->redraw_reason = IPLAT_NULL;
		__is_drawing = 0;
		fwplat->ctl.toolbar_refresh();
		fwplat->ctl.statusbar_refresh();
	}
}

#ifdef PLATFORM_GDI
	int __txt_metric(int swHgt){
		#ifdef PLATFORM_GDI
			int n = 0;
			TEXTMETRIC tm;
			if(fwplat->gui.sysdata && mdc.bmpdc){
				tm.tmHeight = 0;
				tm.tmAveCharWidth = 0;
				dllGetTextMetrics(mdc.bmpdc, &tm);
				if(swHgt)
					n = tm.tmHeight;
				else
					n = tm.tmAveCharWidth;
			}
			return n;
		#endif
	}
#endif

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** FONT FUNCTIONS
 **
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
int 		__impl_winSwFontSelected = 0;
LOGFONTA		lfStock, lfSelected;
void __get_stock_font_info(IPLAT ifont, char **name, int *sze, LOGFONT **lf){
	HGDIOBJ 	obj = 0;
	int			n = sizeof(lfStock);
	memset(&lfStock, 0, n);
	//ANSI_FIXED_FONT, ANSI_VAR_FONT, DEVICE_DEFAULT_FONT, DEFAULT_GUI_FONT, OEM_FIXED_FONT, SYSTEM_FONT, SYSTEM_FIXED_FONT
	if(ifont == FONT_SELECTED && __impl_winSwFontSelected){
		lfStock = lfSelected;
	} else if(ifont == FONT_MONOSPACE){
		obj = GetStockObject(SYSTEM_FIXED_FONT);
		GetObject(obj, n, &lfStock);
	} else {
		obj = GetStockObject(SYSTEM_FONT);
		GetObject(obj, n, &lfStock);
	}

	if(name) *name = lfStock.lfFaceName;
	if(sze){
		if(lfStock.lfHeight == 0)
			*sze = 12;
		if(lfStock.lfHeight < 0)
			*sze = -MulDiv(lfStock.lfHeight, GetDeviceCaps(GetDC(hwndMain), LOGPIXELSY), 72) / 2;
		else
			*sze = lfStock.lfHeight;
	}
	if(lf) *lf = &lfStock;
}

void windows_setfont_execute(){
	#ifdef PLATFORM_GDI_PLUS
		GpFontFamily *ffam = 0;
		INT fstyle = 0,
			style = FontStyleRegular | (__sw_bold ? FontStyleBold : 0) | (__sw_italic ? FontStyleItalic : 0) | (__sw_underlined ? FontStyleUnderline : 0);
		REAL fsize = 0;

		if(Font){
			GdipGetFontStyle(Font, &fstyle);
			if(fstyle != style){
				GdipGetFontSize(Font, &fsize);
				GdipGetFamily(Font, &ffam);

				GdipDeleteFont(Font);
				Font = 0;
				GdipCreateFont(ffam, fsize, style, UnitPoint, &Font);
			}
		}
	#endif
}

void font_use(IPLAT ifont){

	#ifdef PLATFORM_GDI
	LOGFONT *lf = 0;
		if(!fwplat->gui.sysdata || !mdc.windc) return;
		__get_stock_font_info(ifont, 0, 0, &lf);
		if(OldFont) dllSelectObject(mdc.bmpdc, OldFont);
		OldFont = 0;
		if(Font) dllDeleteObject(Font);
		Font = CreateFontIndirect(lf);
		if(Font){
			OldFont = (HFONT)dllSelectObject(mdc.bmpdc, Font);
			__caret_hgt =  fwplat->gui.fonthgt();
			__caret_wdt = fwplat->gui.fontwdt();
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		GpFontFamily *ffam = 0;

		if(Font) GdipDeleteFont(Font);
		Font = 0;

		if(ifont == FONT_SELECTED && __impl_winSwFontSelected){
			char *fface;
			__get_stock_font_info(ifont, &fface, &__def_font_sze, 0);
			fwplat->gui.font_face(fface, __def_font_sze);
		} else if(ifont == FONT_MONOSPACE){
			GdipGetGenericFontFamilyMonospace(&ffam);
			GdipCreateFont(ffam, __def_font_sze, FontStyleRegular, UnitPoint, &Font);
		} else {
			GdipGetGenericFontFamilySerif(&ffam);
			GdipCreateFont(ffam, __def_font_sze, FontStyleRegular, UnitPoint, &Font);
		}

		fwplat->gui.lenpixels(&__font_wdt, &__font_hgt, "X");
	#endif

	#ifdef PLATFORM_DIRECT2D
		char *fface;
		__get_stock_font_info(ifont, &fface, &__def_font_sze, 0);
		dllDX2D_usefont(fface, __def_font_sze, __sw_bold, __sw_italic);
	#endif
}

int font_face(const char *face, int sze){
	#ifdef PLATFORM_GDI
		char fface[1024 * 4];
		if(!fwplat->gui.sysdata || !mdc.windc) return 0;

		if(!face || !strlen(face))
			dllGetTextFace(mdc.windc, TXTOUT_MAX, fface);
		else
			sprintf(fface, "%s", face);

		if(sze <= 0) sze = __def_font_sze;

		if(OldFont) dllSelectObject(mdc.bmpdc, OldFont);
		OldFont = 0;
		if(Font) dllDeleteObject(Font);
		Font = (HFONT)dllCreateFont(sze, sze, 0, 0, FW_DONTCARE, 0, 0, 0, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, fface);
		if(Font){
			OldFont = (HFONT)dllSelectObject(mdc.bmpdc, Font);
			__caret_hgt =  fwplat->gui.fonthgt();
			__caret_wdt = fwplat->gui.fontwdt();
		} else {
			font_use(0);
			return 0;
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		GpFontFamily *ffam = 0;

		if(!face || !strlen(face) ) return 0;
		if(sze <= 0) sze = __def_font_sze;

		GdipCreateFontFamilyFromName(WC(face, 0), NULL, &ffam);
		if(ffam){
			if(Font) GdipDeleteFont(Font);
			Font = 0;
			GdipCreateFont(ffam, sze, FontStyleRegular, UnitPoint, &Font);
			GdipDeleteFontFamily(ffam);
			fwplat->gui.lenpixels(&__font_wdt, &__font_hgt, "X");
		}
		if(!Font){
			font_use(0);
			return 0;
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
		if(!SUCCEEDED(dllDX2D_usefont(face, sze, __sw_bold, __sw_italic))){
			font_use(0);
			return 0;
		}
	#endif

	return 1;	//succeeded
}

int fontwdt(void){
	#ifdef PLATFORM_GDI
		return __txt_metric(0);
	#endif

	#ifdef PLATFORM_GDI_PLUS
		return __font_wdt;
	#endif

	#ifdef PLATFORM_DIRECT2D
		return dllDX2D_fontw();
	#endif
}

int fonthgt(void){
	#ifdef PLATFORM_GDI
		return __txt_metric(1);
	#endif

	#ifdef PLATFORM_GDI_PLUS
		return __font_hgt;
	#endif

	#ifdef PLATFORM_DIRECT2D
		return dllDX2D_fonth();
	#endif
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** COLOR FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
void set_alpha(double na){
	__alpha = CLAMP(na, 0.0, 1.0);
}

void set_intensity(double ni){
	__intensity = CLAMP(ni, 0.0, 1.0);
}

void set_fcolor(int r, int g, int b){
	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
			fcol = RGB(r * __intensity, g * __intensity, b * __intensity);
			if(OldPen) dllSelectObject(mdc.bmpdc, OldPen);
			OldPen = 0;
			if(Pen) dllDeleteObject(Pen);

			Pen = (HPEN)dllCreatePen(PS_SOLID, 0, fcol);
			if(Pen) OldPen = (HPEN)dllSelectObject(mdc.bmpdc, Pen);

			dllSetTextColor(mdc.bmpdc, fcol);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		ARGB col = MAKE_ARGB(r * __intensity, g * __intensity, b * __intensity, __alpha * col_maxval);
		if(!Pen) 	GdipCreatePen1(col, 1.f, UnitPixel, &Pen);
		else 		GdipSetPenColor(Pen, col);
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_color_pen(col_maxval, r * __intensity, g * __intensity, b * __intensity, __alpha);
	#endif
}

void set_bcolor(int r, int g, int b){
	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
			bcol = RGB(r * __intensity, g * __intensity, b * __intensity);
			if(OldBrush) dllSelectObject(mdc.bmpdc, OldBrush);
			OldBrush = 0;
			if(Brush) dllDeleteObject(Brush);

			Brush = (HBRUSH)dllCreateSolidBrush(bcol);
			if(Brush) OldBrush = (HBRUSH)dllSelectObject(mdc.bmpdc, Brush);

			dllSetBkColor(mdc.bmpdc, bcol);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		ARGB col = MAKE_ARGB(r * __intensity, g * __intensity, b * __intensity, __alpha * col_maxval);
		if(!Brush) 	GdipCreateSolidFill(col, &Brush);
		else		GdipSetSolidFillColor(Brush, col);
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_color_brush(col_maxval, r * __intensity, g * __intensity, b * __intensity, __alpha);
	#endif
}

void __impl_set_fcolorn(double nr, double ng, double nb){
	colorf.r = CLAMP(nr, 0.0, 1.0) * col_maxval;
	colorf.g = CLAMP(ng, 0.0, 1.0) * col_maxval;
	colorf.b = CLAMP(nb, 0.0, 1.0) * col_maxval;
	set_fcolor(colorf.r, colorf.g, colorf.b);
}
void __impl_set_bcolorn(double nr, double ng, double nb){
	colorb;
	colorb.r = CLAMP(nr, 0.0, 1.0) * col_maxval;
	colorb.g = CLAMP(ng, 0.0, 1.0) * col_maxval;
	colorb.b = CLAMP(nb, 0.0, 1.0) * col_maxval;
	set_bcolor(colorb.r, colorb.g, colorb.b);
}

void set_sys_color(IPLAT f, IPLAT b){
	double	dr, dg, db, da;
	C_RGBA 	F, B,
			*fg = &color_fg,
			*bg = &color_bg;
	#ifndef PLATFORM_DIRECT2D
	if(fwplat->gui.sysdata && mdc.bmpdc){
	#endif
		/*foreground*/
		if(f >= SYSCOL_BLACK && f <= SYSCOL_CYANL){
			plat_normal_color(f, &dr, &dg, &db, &da);
			F.r = dr * col_maxval;
			F.g = dg * col_maxval;
			F.b = db * col_maxval;
			F.a = da * col_maxval;
			fg = &F;
		} else if(f == SYSCOL_TEXT_F)		fg = &color_fg;
		else if(f == SYSCOL_TEXT_B)			fg = &color_bg;
		else if(f == SYSCOL_TEXT_SELETED_F)	fg = &color_sel_fg;
		else if(f == SYSCOL_TEXT_SELETED_B)	fg = &color_sel_bg;
		else if(f == SYSCOL_TOOLTIP_F)		fg = &color_tip_fg;
		else if(f == SYSCOL_TOOLTIP_B)		fg = &color_tip_bg;
		else if(f == SYSCOL_BUTTON_F)		fg = &color_btn_fg;
		else if(f == SYSCOL_BUTTON_B)		fg = &color_btn_bg;

		/*background*/
		if(b >= SYSCOL_BLACK && b <= SYSCOL_CYANL){
			plat_normal_color(b, &dr, &dg, &db, &da);
			B.r = dr * col_maxval;
			B.g = dg * col_maxval;
			B.b = db * col_maxval;
			B.a = da * col_maxval;
			bg = &B;
		} else if(b == SYSCOL_TEXT_F)		bg = &color_fg;
		else if(b == SYSCOL_TEXT_B)			bg = &color_bg;
		else if(b == SYSCOL_TEXT_SELETED_F)	bg = &color_sel_fg;
		else if(b == SYSCOL_TEXT_SELETED_B)	bg = &color_sel_bg;
		else if(b == SYSCOL_TOOLTIP_F)		bg = &color_tip_fg;
		else if(b == SYSCOL_TOOLTIP_B)		bg = &color_tip_bg;
		else if(b == SYSCOL_BUTTON_F)		bg = &color_btn_fg;
		else if(b == SYSCOL_BUTTON_B)		bg = &color_btn_bg;

		colorf = *fg;
		colorb = *bg;

		set_fcolor(colorf.r, colorf.g, colorf.b);
		set_bcolor(colorb.r, colorb.g, colorb.b);

	#ifndef PLATFORM_DIRECT2D
	}
	#endif
}

void __imp_swap_colors(void){
	C_RGBA	color = colorf;

	colorf = colorb;
	colorb = color;

	#ifdef PLATFORM_GDI_PLUS
		set_fcolor(colorf.r, colorf.g, colorf.b);
		set_bcolor(colorb.r, colorb.g, colorb.b);
	#endif
}
/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** DRAWING FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
void clrscr(void){
	#ifdef PLATFORM_GDI
		RECT r;
		if(fwplat->gui.sysdata && mdc.bmpdc){
			GetClientRect(hwndChild, &r);
			dllRectangle(mdc.bmpdc, r.left-1, r.top-1, r.right+1, r.bottom+1);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			ARGB col = MAKE_ARGB(0, 0, 0 , 255);
			GdipGetSolidFillColor(Brush, &col);
			GdipGraphicsClear(gdpGraphics, col);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_cls();
	#endif
}

void putpixel(int x, int y){
	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
			dllSetPixel(mdc.bmpdc, x, y, fcol);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			GdipDrawRectangle(gdpGraphics, Pen, x, y, x+1, y+1);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_rectangle(0, x, y, 1, 1);
	#endif
}

void line(int x1, int y1, int x2, int y2){
	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
			dllMoveToEx(mdc.bmpdc, x1, y1, 0);
			dllLineTo(mdc.bmpdc, x2, y2);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			GdipDrawLine(gdpGraphics, Pen, x1, y1, x2, y2);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_line(x1, y1, x2, y2);
	#endif
}

void dashline(int x1, int y1, int x2, int y2){
	#ifdef PLATFORM_GDI
		HPEN DashPen = 0, CurPen = 0;
		if(fwplat->gui.sysdata && mdc.bmpdc){
			DashPen = (HPEN)dllCreatePen(PS_DASH, 0, fcol);
			if(DashPen) CurPen = (HPEN)dllSelectObject(mdc.bmpdc, DashPen);
			line(x1, y1, x2, y2);
			if(CurPen) dllSelectObject(mdc.bmpdc, CurPen);
			if(DashPen) dllDeleteObject(DashPen);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			GdipSetPenDashStyle(Pen, DashStyleDash);
			GdipDrawLine(gdpGraphics, Pen, x1, y1, x2, y2);
			GdipSetPenDashStyle(Pen, DashStyleSolid);
		}
	#endif
	#ifdef PLATFORM_DIRECT2D
	dllDX2D_dashline(x1, y1, x2, y2);
	#endif
}

void winrectangle(int swFilled, int x, int y, int wdt, int hgt){
	#ifdef PLATFORM_GDI
		RECT r;
		if(fwplat->gui.sysdata && mdc.bmpdc){
			r.left = x;
			r.top = y;
			r.right = x + wdt;
			r.bottom = y + hgt;
			if(swFilled){
				dllFillRect(mdc.bmpdc, &r, Brush);
			} else {
				fwplat->gui.line(x, y, x + wdt, y);
				fwplat->gui.line(x + wdt, y, x + wdt, y + hgt);
				fwplat->gui.line(x, y + hgt, x + wdt, y + hgt);
				fwplat->gui.line(x, y, x, y + hgt);
			}
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			if(swFilled)
				GdipFillRectangle(gdpGraphics, Brush, x, y, wdt, hgt);
			else
				GdipDrawRectangle(gdpGraphics, Pen, x, y, wdt, hgt);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_rectangle(swFilled, x, y, wdt, hgt);
	#endif
}

void wincircle(int swFilled, int x, int y, int r){

	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
		    if(swFilled){
			    r /= 2;
		    	dllEllipse(mdc.bmpdc, x - r, y - r, x + r, y + r);
		    } else {
				dllArc(mdc.bmpdc, 	x - r, y - r,
									x + r, y + r,
									x + cos(0) * r, y + sin(0) * r,
									x + cos(M_PI * -2) * r, y + sin(M_PI * -2) * r);
		    }
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics){
			if(swFilled)
				GdipFillEllipse(gdpGraphics, Brush, x - r, y - r, r * 2, r * 2);
			else
				GdipDrawEllipse(gdpGraphics, Pen, x - r, y - r, r * 2, r * 2);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_circle(swFilled, x, y, r);
	#endif
}

#ifdef PLATFORM_GDI_PLUS
const GpFillMode pollyFill = FillModeWinding;
//const GpFillMode pollyFill = FillModeAlternate;
#endif

void winpolygon(int swFilled, int elements, double *x, double *y){
	#ifdef PLATFORM_GDI
		POINT *lpPoints = 0;
		int p;
		if(fwplat->gui.sysdata && mdc.bmpdc){
			lpPoints = malloc(sizeof(POINT) * elements);
			if(!lpPoints) return;
			for(p = 0; p < elements; p++){
				lpPoints[p].x = (LONG)(.5f + x[p]);
				lpPoints[p].y = (LONG)(.5f + y[p]);
			}
			if(swFilled){
				dllPolygon(mdc.bmpdc, lpPoints, elements);
			}else
				dllPolyline(mdc.bmpdc, lpPoints, elements);

			free(lpPoints);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		PointF *lpPoints = 0;
		int p;

		if(gdpGraphics){
			lpPoints = malloc(sizeof(PointF) * elements);
			if(!lpPoints) return;
			for(p = 0; p < elements; p++){
				lpPoints[p].X = (REAL)(x[p]);
				lpPoints[p].Y = (REAL)(y[p]);
			}

			if(swFilled)
				GdipFillPolygon(gdpGraphics, Brush, lpPoints, elements, pollyFill);
			else
				GdipDrawPolygon(gdpGraphics, Pen, lpPoints, elements);
			free(lpPoints);
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
	dllDX2D_polygon(swFilled, elements, x, y);
	#endif
}

void winlinepath(int elements, double *x, double *y){
	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc && elements > 1){
			dllMoveToEx(mdc.bmpdc, x[0], y[0], 0);
			for(int i = 1; i < elements; i++){
				dllLineTo(mdc.bmpdc, x[i], y[i]);
			}
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics && elements > 1){
			for(int i = 0; i < elements - 1; i++){
				GdipDrawLine(gdpGraphics, Pen, x[i], y[i], x[i + 1], y[i + 1]);
			}
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
		dllDX2D_linepath(elements, x, y);
	#endif

}

void textmode(IPLAT text_mode){
	if(text_mode == TEXT_OPAQUE)
		__sw_opaque = 1;
	else if(text_mode == TEXT_TRANSPARENT)
		__sw_opaque = 0;
	else if(text_mode == TEXT_NORMAL)
		__sw_bold = __sw_italic = __sw_underlined = 0;
	else if(text_mode == TEXT_BOLD)
		__sw_bold = 1;
	else if(text_mode == TEXT_ITALIC)
		__sw_italic = 0;
	else if(text_mode == TEXT_UNDERLINED)
		__sw_underlined = 0;

	windows_setfont_execute();
	#ifdef PLATFORM_DIRECT2D
	dllDX2D_set_text(__sw_bold, __sw_italic, __sw_opaque);
	#endif
}


/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** TEXT FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
#ifdef PLATFORM_GDI_PLUS
	GpStringFormat* __string_formatter(void){
		if(gdpSF) return gdpSF;

		GdipStringFormatGetGenericDefault(&gdpSF);
		GdipSetStringFormatFlags(gdpSF, 	StringFormatFlagsNoWrap |
											StringFormatFlagsNoFitBlackBox |
											StringFormatFlagsNoFontFallback |
											StringFormatFlagsMeasureTrailingSpaces |
											StringFormatFlagsNoWrap |
											StringFormatFlagsNoClip
		);
		return gdpSF;
	}
#endif

int __string_metrics(char *str, int *wdt, int *hgt){
	WCHAR 	*wc = WC(str, 0);
	int		len = wcslen(wc);
	if(wdt) *wdt = 0;
	if(hgt) *hgt = 0;
	if(len){

		#ifdef PLATFORM_GDI
			SIZE metric;
			if(fwplat->gui.sysdata && mdc.bmpdc && str){
				metric.cx = metric.cy = 0;
				dllGetTextExtentPointW(mdc.bmpdc, wc, len, &metric);
				if(wdt) *wdt = metric.cx;
				if(hgt) *hgt = metric.cy;
			}
		#endif

		#ifdef PLATFORM_GDI_PLUS
			RectF ra, rb;
			GpStringFormat *sf = 0;

			if(Font && gdpGraphics){
				ra.X = ra.Y = ra.Height = ra.Width = 0;

				sf = __string_formatter();
				GdipMeasureString(gdpGraphics, wc, len, Font, &ra, sf, &rb, NULL, NULL);

				if(wdt) *wdt = rb.Width;
				if(hgt) *hgt = rb.Height;
			}
		#endif

		#ifdef PLATFORM_DIRECT2D
			dllDX2D_lentextw(wc, len, wdt, hgt);
		#endif
	}

	return len;
}

int cpixels(int *wdt, int *hgt, char *str){
	return __string_metrics(str, wdt, hgt);
}

void cout(int x, int y, char *str){
	WCHAR 	*wc = WC(str, 0);
	int		len = wcslen(wc);
	if(!len) return;


	#ifdef PLATFORM_GDI
		if(fwplat->gui.sysdata && mdc.bmpdc){
			set_fcolor(colorf.r, colorf.g, colorf.b);
			set_bcolor(colorb.r, colorb.g, colorb.b);
			dllSetBkMode( mdc.bmpdc, (__sw_opaque ? OPAQUE : TRANSPARENT) );
			dllTextOutW(mdc.bmpdc, x, y, wc, len);
		}
	#endif

	#ifdef PLATFORM_GDI_PLUS
		if(gdpGraphics == 0) return;
		ARGB fcol, bcol;

		GdipGetPenColor(Pen, &fcol);
		GdipGetSolidFillColor(Brush, &bcol);

		GpStringFormat *sf = __string_formatter();
		RectF ra, rb;

		ra.X = x; ra.Y = y; ra.Height = ra.Width = 0;
		GdipMeasureString(gdpGraphics, wc, len, Font, &ra, sf, &rb, NULL, NULL);


		if(__sw_opaque){
			if(__sw_reverse)
				GdipSetSolidFillColor(Brush, fcol);
			else
				GdipSetSolidFillColor(Brush, bcol);

			GdipFillRectangle(gdpGraphics, Brush, rb.X, rb.Y, rb.Width, rb.Height);
		}

		GdipSetSolidFillColor(Brush, fcol);

		GdipDrawString(gdpGraphics, wc, len, Font, &rb, sf, Brush);

		//restore the brush to her original color
		GdipSetSolidFillColor(Brush, bcol);
	#endif

	#ifdef PLATFORM_DIRECT2D
		set_fcolor(colorf.r, colorf.g, colorf.b);
		set_bcolor(colorb.r, colorb.g, colorb.b);
		dllDX2D_outtextw(x, y, wc, len);
	#endif

}

void textoutva(int x, int y, char *fmt, va_list args){
	char 	buf[TXTOUT_MAX];
	if(fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		cout(x, y, buf);
	}
}

void textout(int x, int y, char *fmt, ...){
	va_list args;
	if(fmt){
    	va_start(args, fmt);
    	textoutva(x, y, fmt, args);
    	va_end(args);
	}
}

int lenpixelsva(int *wdt, int *hgt, char *fmt, va_list args){
	char buf[TXTOUT_MAX];
	if(fmt){
		vsnprintf(buf, TXTOUT_MAX, fmt, args);
		__string_metrics(buf, wdt, hgt);
	}
	return 0;
}

int lenpixels(int *wdt, int *hgt, char *fmt, ...){
	va_list args;
	int stat = 0;
	if(fmt){
    	va_start(args, fmt);
    	stat = lenpixelsva(wdt, hgt, fmt, args);
    	va_end(args);
	}
	return stat;
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** CARET FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
void __create_caret(int swOn){
	if(hwndChild){
		if(__caret_created) DestroyCaret();
		if(swOn){
			CreateCaret(hwndChild, (HBITMAP)0, __caret_mode == CARET_OVW ? __caret_wdt : 2, __caret_hgt + __caret_gap);
			ShowCaret(hwndChild);
		}
		__caret_created = swOn;
	}
}

void caret_set(IPLAT caret_onoff, IPLAT caret_mode, int x, int y, char letter){
	__caret_mode = caret_mode;
	POINT p;
	GetCaretPos(&p);
	if(hwndMain){
		if(!__sw_caret_moved)
			__sw_caret_moved = (p.x == x && p.y == y);
		if(caret_onoff == CARET_OFF){
			__create_caret(0);
			__sw_showCaret = 0;
		} else {
			__create_caret(__got_focus);
			SetCaretPos(x, y);
			__sw_showCaret = 1;
		}
	}
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** WINDOW FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
int maxx(void){
	RECT r;
	GetClientRect(hwndChild, &r);
	return (int)(r.right - r.left);
}
int maxy(void){
	RECT r;
	GetClientRect(hwndChild, &r);
	return (int)(r.bottom - r.top);
}

int resolutionx(void){
	RECT r;
	GetClientRect(GetDesktopWindow(), &r);
	return (int)(r.right - r.left);
}
int resolutiony(void){
	RECT r;
	GetClientRect(GetDesktopWindow(), &r);
	return (int)(r.bottom - r.top);
}

void doevents(){
	MSG msg;
	while(!swWinExit && PeekMessageW(&msg, NULL, 0, 0, PM_REMOVE)){
		GetMessageW( &msg, 0, 0, 0);
		TranslateMessage(&msg);
		DispatchMessageW(&msg);
	}
}

void* screencopy(int x, int y, int wdt, int hgt){
	HBITMAP	bmpshot = 0;
	HDC dcshot = 0;

	if(fwplat->gui.sysdata && mdc.bmpdc && mdc.windc){
    	dcshot = (HDC)dllCreateCompatibleDC(mdc.windc);
    	if(dcshot){
    		bmpshot = (HBITMAP)dllCreateCompatibleBitmap(mdc.bmpdc, wdt, hgt);
    		if(bmpshot){
    	    	dllSelectObject(dcshot, bmpshot);
    			if( !dllBitBlt(dcshot, 0, 0, wdt, hgt, mdc.bmpdc, x, y, SRCCOPY) ){
    				dllDeleteObject((HBITMAP)bmpshot);
    				bmpshot = 0;
    			}
    		}
    	   	dllDeleteObject(dcshot);
    	}
	}

	return bmpshot;
}

void screenput(void *bmpshot, int scrx, int scry, int imgx, int imgy, int wdt, int hgt){
	HDC dcshot = 0;
	if(bmpshot && fwplat->gui.sysdata && mdc.bmpdc && mdc.windc){
		dcshot = (HDC)dllCreateCompatibleDC(mdc.windc);
    	if(dcshot){
        	dllSelectObject(dcshot, (HBITMAP)bmpshot);
    		dllBitBlt(mdc.bmpdc, scrx, scry, wdt, hgt, dcshot, imgx, imgy, SRCCOPY);
    	}
	   	dllDeleteObject(dcshot);
	}
}
void* screendelete(void *bmpshot){
	if(bmpshot) dllDeleteObject((HBITMAP)bmpshot);
	return 0;
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** TIME FUNCTIONS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
struct timeval __impl_time(){
	struct timeval tv;
	if(gettimeofday(&tv, 0) == -1){
		tv.tv_sec = 0;
		tv.tv_usec = 0;
	}
	return tv;
}

#define FSEC 1.0E-6
#define USEC 1.0E+6
float __impl_time_delta(struct timeval tv){
	struct timeval now = __impl_time();
	float 	sec = (now.tv_sec - tv.tv_sec - 1),
			usec = FSEC * (USEC + now.tv_usec - tv.tv_usec);
	//printf("sec %.3f, usec %.3f = %.3f\n", sec, usec, sec + usec);
	return sec + usec;
}

void __kill_timers(void){
	thread_data_t *ti;

	if(__timer_handler){
		KillTimer(hwndMain, __timer_handler);
		__timer_handler = 0;
	}
	if(lstTimer){
		while(lstTimer->cnt){
			ti = (thread_data_t*)list_pop(lstTimer);
			if(ti) free(ti);
		};
		lstTimer = list_destroy(lstTimer);
	}
}

int __sw_in_timer = 0;
void CALLBACK __proc_timer(HWND wnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime){
	LIST_INDEX 		i;
	thread_data_t 	*ti;
	float			dt;

	if(__sw_in_timer || !lstTimer) return;
	__sw_in_timer = 1;
	for(i = 0; i < lstTimer->cnt; i++){
		ti = (thread_data_t*)list_at(lstTimer, i);
		if(!ti) continue;

		dt = __impl_time_delta(ti->tv);

		if(dt >= ti->msec){
			ti->tv = __impl_time();
			if(!ti->tmfunc(ti->tdata)){
				free(ti);
				lstTimer = list_del(lstTimer, i);
				i--;
			}
		}
	}
	__sw_in_timer = 0;
}

int timed_call(int msec, __timer_routine tr, void *arg){
	thread_data_t *ti;
	if(!tr) return 0;
	if(!__timer_handler)__timer_handler = SetTimer(hwndMain, 0, USER_TIMER_MINIMUM, __proc_timer);
	if(!__timer_handler) return 0;

	if(!lstTimer){
		lstTimer = list_new(LIST_MIN_BLOCK);
		if(!lstTimer) return 0;
	}

	ti = (thread_data_t*)malloc(sizeof(thread_data_t));
	if(!ti) return 0;
	memset(ti, 0, sizeof(thread_data_t));
	ti->tmfunc = tr;
	ti->tdata = arg;
	ti->msec = msec;
	ti->msec = ti->msec / 1000.f;
	ti->tv = __impl_time();
	lstTimer = list_add(lstTimer, LIST_ELEMENT_CAST(ti));
	return 1;
}

void win_impl_exit(int exitcode){
	PostQuitMessage(exitcode);
}

double total_memory(void){
	double mb = 0;
	MEMORYSTATUS mstat;
	_local_MEMORYSTATUSEX mstatex;

	if(dllGlobalMemoryStatusEx){
		mstatex.dwLength = sizeof(mstatex);
		dllGlobalMemoryStatusEx(&mstatex);
		mb = mstatex.ullTotalPhys / (double)(1024 * 1024);
	} else {
		mstat.dwLength = sizeof(mstat);
		GlobalMemoryStatus(&mstat);
		mb = (double)mstat.dwTotalPhys / (double)(1024 * 1024);
	}
	return mb;
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** SCROLLBARS
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
void sbar_set(IPLAT sbar, double pct){
	SCROLLINFO si;
	double val, ds;
	if(sbar == SBARV){
		memset(&si, 0, sizeof(SCROLLINFO));
		si.cbSize = sizeof(SCROLLINFO);
		si.fMask = SIF_POS | SIF_RANGE;
		GetScrollInfo(hwndChild, SB_VERT, &si);
		ds = (double)(si.nMax - si.nMin);
		val = ds / 100.f * pct;
		si.nPos = (int)val;
		SetScrollInfo(hwndChild, SB_VERT, &si, 1);
	} else if(sbar == SBARH){
		memset(&si, 0, sizeof(SCROLLINFO));
		si.cbSize = sizeof(SCROLLINFO);
		si.fMask = SIF_POS | SIF_RANGE;
		GetScrollInfo(hwndChild, SB_HORZ, &si);
		ds = (double)(si.nMax - si.nMin);
		val = ds / 100.f * pct;
		si.nPos = (int)val;
		SetScrollInfo(hwndChild, SB_HORZ, &si, 1);
	}
}
void on_srollbar_changed(UINT sbar, double pct){
	if(fwplat && !__is_drawing){
		if(sbar & SB_HORZ){
			onscrollbar(fwplat, SBARH, pct);
		}
		if(sbar & SB_VERT){
			onscrollbar(fwplat, SBARV, pct);
		}
	}
}


void sbar_show(IPLAT sbar, int sw){
	if(sbar == SBARH){
		EnableScrollBar(hwndChild, SB_HORZ, ESB_ENABLE_BOTH);
		ShowScrollBar(hwndChild, SB_HORZ, sw);
	}  else if(sbar == SBARV){
		EnableScrollBar(hwndChild, SB_VERT, ESB_ENABLE_BOTH);
		ShowScrollBar(hwndChild, SB_VERT, sw);
	}
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** MAIN WINDOW
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
void resize_children(){
	RECT r, rtmp;
	int h, hgtStatusBar = 0;
	GetClientRect(hwndMain, &r);
	h = r.bottom;
	if(hwndToolBar){
		if(lstToolBarItems->cnt){
			GetWindowRect(hwndToolBar, &rtmp); //<------<< only to get the height of the toolbar control, GetClientRect returns wrong dimensions
			rtmp.bottom -= rtmp.top;
			r.bottom = r.top + rtmp.bottom;
			SetWindowPos(hwndToolBar, HWND_TOP, r.left, 0, r.right, r.top  - r.top, SWP_SHOWWINDOW);
			r.top = r.bottom;
		} else if(IsWindowVisible(hwndToolBar)){
			ShowWindow(hwndToolBar, SW_HIDE);
		}
	}

	if(hwndTab){
		if(lstTabPages->cnt){
			r.bottom = r.top + TabHeight;
			SetWindowPos(hwndTab, HWND_TOP, r.left, r.top, r.right - TabHeight, r.bottom - r.top, SWP_SHOWWINDOW);
			SetWindowPos(hwndTabCloseBtn, HWND_TOP, r.right - TabHeight, r.top, TabHeight, TabHeight, SWP_SHOWWINDOW);
			r.top = r.bottom;
		} else if(IsWindowVisible(hwndTab)){
			ShowWindow(hwndTab, SW_HIDE);
			ShowWindow(hwndTabCloseBtn, SW_HIDE);
		}
	}

	if(lstStatusBar->cnt){
		GetWindowRect(hwndStatus, &rtmp); //<------<< only to get the height of the toolbar control, GetClientRect returns wrong dimensions
		hgtStatusBar = rtmp.bottom - rtmp.top;
	} else
		ShowWindow(hwndStatus, SW_HIDE);

	if(hwndChild){
		r.bottom = h - r.top - hgtStatusBar;
		SetWindowPos(hwndChild, HWND_TOP, r.left, r.top, r.right, r.bottom, SWP_SHOWWINDOW);
		r.top = r.bottom;
	}

	if(lstStatusBar->cnt){
		r.bottom = h;
		SetWindowPos(hwndStatus, HWND_TOP, r.left, r.top, r.right, r.bottom, SWP_SHOWWINDOW);
	}
}

int MB_LEFT_IN = 0;
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam){
	PLATKEY 	pk = PKAN;
	TCITEM 		tie;
	SCROLLINFO 	si;
	double		pct, d;
	LPNMMOUSE	lpnm;
	LPNMHDR		nphdr;
	void		*p;
	int			i;
	PAINTSTRUCT ps;
	HDC			hdc;
	PLAT_RECT	r;

	CTL_TOOLBAR_ITEM *tbitem;
		switch (nMsg) {
		case WM_NOTIFY:
			switch (((LPNMHDR) lParam)->code){
			case NM_CLICK: /*toolbar button clicked*/
				lpnm = (LPNMMOUSE) lParam;
				nphdr = (LPNMHDR)&lpnm->hdr;
				if(nphdr->hwndFrom == hwndToolBar){
					tbitem = (CTL_TOOLBAR_ITEM*)(lpnm->dwItemData);
					on_toolbar_item_clicked(tbitem);
				} else if(nphdr->hwndFrom == hwndTab){
					memset(&tie, 0, sizeof(TCITEM));
					tie.mask = TCIF_TEXT | TCIF_PARAM;
					i = TabCtrl_GetCurSel(hwndTab);
					if(TabCtrl_GetItem(hwndTab, i, &tie)){
						p = (void*)tie.lParam;
						fwplat->redraw_reason = REDRAW_MENU_ACTION;
						ontabactive(fwplat, i + 1);
						__win_redraw(0);
					}
				}
				break;
			case TTN_GETDISPINFO:
				break;
			default:
				break;
			}

			break;

		case WM_VSCROLL:
			if(LOWORD(wParam) == SB_TOP){
			} else if(LOWORD(wParam) == SB_BOTTOM){
			} else if(LOWORD(wParam) == SB_LINEUP){
				onskey(fwplat, PKUP, 1);
			} else if(LOWORD(wParam) == SB_PAGEUP){
				onskey(fwplat, PKPU, 1);
			} else if(LOWORD(wParam) == SB_PAGEDOWN){
				onskey(fwplat, PKPD, 1);
			} else if(LOWORD(wParam) == SB_LINEDOWN){
				onskey(fwplat, PKDN, 1);
			} else if(LOWORD(wParam) == SB_THUMBTRACK){
				memset(&si, 0, sizeof(SCROLLINFO));
				si.cbSize = sizeof(SCROLLINFO);
				si.fMask = SIF_TRACKPOS | SIF_RANGE;
				GetScrollInfo(hwndChild, SB_VERT, &si);
				d = (double)(si.nMax - si.nMin);
				pct = (double)(si.nTrackPos) / (d ? d : 1.f) * 100.f;
				on_srollbar_changed(SB_VERT, pct);
				si.fMask = SIF_POS;
				si.nPos = si.nTrackPos;
				si.fMask = SIF_TRACKPOS;
				SetScrollInfo(hwndChild, SB_VERT, &si, 0);
			}
			break;

		case WM_HSCROLL:
			if(LOWORD(wParam) == SB_TOP){
			} else if(LOWORD(wParam) == SB_BOTTOM){
			} else if(LOWORD(wParam) == SB_LINEUP){
				onskey(fwplat, PKLE, 1);
			} else if(LOWORD(wParam) == SB_PAGEUP){
				onskey(fwplat, PKPL, 1);
			} else if(LOWORD(wParam) == SB_PAGEDOWN){
				onskey(fwplat, PKPR, 1);
			} else if(LOWORD(wParam) == SB_LINEDOWN){
				onskey(fwplat, PKRI, 1);
			} else if(LOWORD(wParam) == SB_THUMBTRACK){
				memset(&si, 0, sizeof(SCROLLINFO));
				si.cbSize = sizeof(SCROLLINFO);
				si.fMask = SIF_TRACKPOS | SIF_RANGE;
				GetScrollInfo(hwndChild, SB_HORZ, &si);
				d = (double)(si.nMax - si.nMin);
				pct = (double)(si.nTrackPos) / (d ? d : 1.f) * 100.f;
				on_srollbar_changed(SB_HORZ, pct);
				si.fMask = SIF_POS;
				si.nPos = si.nTrackPos;
				si.fMask = SIF_TRACKPOS;
				SetScrollInfo(hwndChild, SB_HORZ, &si, 0);
			}
			break;

		case  WM_SIZE:
			fwplat->redraw_reason = REDRAW_RESIZE;
			if(hwnd == hwndChild){
				if(sbarflags){
					memset(&si, 0, sizeof(SCROLLINFO));
					si.cbSize = sizeof(SCROLLINFO);
					si.fMask = SIF_RANGE;
					si.nMin = 0;
					if(sbarflags & SB_VERT){
						si.nMax = HIWORD(lParam);
						SetScrollInfo(hwndChild, SB_VERT, &si, 1);
					}
					if(sbarflags & SB_HORZ){
						si.nMax = LOWORD(lParam);
						SetScrollInfo(hwndChild, SB_HORZ, &si, 1);
					}
				}
			}
			if(hwnd == hwndMain){
				/*set the client areas for the child and tab windows*/
				resize_children();
			}
			break;

		case WM_MOUSEWHEEL:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
			if(GET_WHEEL_DELTA_WPARAM(wParam) > 0)
				onmouse(fwplat, mouseX, mouseY, MWHEEL_UP);
			else
				onmouse(fwplat, mouseX, mouseY, MWHEEL_DN);
			break;

		case WM_LBUTTONUP:
		case WM_NCLBUTTONUP:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
				MB_LEFT_IN = 0;
				onmouse(fwplat, mouseX, mouseY, MBTN_LEFT_UP);
			break;

		case WM_LBUTTONDOWN:
		case WM_NCLBUTTONDOWN:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
				MB_LEFT_IN = 1;
				onmouse(fwplat, mouseX, mouseY, MBTN_LEFT_IN);
			break;

		case WM_NCMOUSEMOVE:
			//is for the tab
			if(!IsWindowVisible(hwndTab))
				return DefWindowProcW(hwnd, nMsg, wParam, lParam);

		case WM_MOUSEMOVE:
			mouseX = LOWORD(lParam);
			mouseY = HIWORD(lParam);
			mouseB = (wParam == MK_LBUTTON ? MBTN_LEFT_IN : IPLAT_NULL);
				onmouse(fwplat, mouseX, mouseY, mouseB);
			break;

		case WM_SETFOCUS:
			if(!__got_focus){
				__got_focus = 1;
				__create_caret(__sw_showCaret);
				__win_redraw(0);
			}
			break;

		case WM_KILLFOCUS:
			if(__got_focus){
				__got_focus = 0;
				__create_caret(0);
				__win_redraw(0);
			}
			break;

		case WM_CLOSE:
			if(fwplat && ongui_exit(fwplat)){
				swWinExit = 1;
				PostQuitMessage(0);
				return 1;
			}
			return 0;
			break;

		case WM_SETTINGCHANGE:
		case WM_DISPLAYCHANGE:
			swUpdateColors = 1;
			break;

		case WM_PAINT:
			//update the screen
			if(hwnd == hwndChild && !__is_drawing){
				hdc = BeginPaint(hwnd, &ps);
				if(hdc){
					PLAT_RECT_SET(r, (int)ps.rcPaint.left, (int)ps.rcPaint.top, (int)ps.rcPaint.right, (int)ps.rcPaint.bottom)
					windows_setfont_execute();
					__begin_draw(hdc);
					onredraw(fwplat, &r);
					__end_draw(&ps);
				}
				EndPaint(hwnd, &ps);
				return 0L;
			}
			break;

		case WM_KEYUP:
		case WM_KEYDOWN:
			kbd = (int)wParam;
			//virtual keys http://msdn.microsoft.com/en-us/library/ms927178.aspx
			if(kbd == VK_DOWN)pk = PKDN;
			else if(kbd == VK_UP)pk = PKUP;
			else if(kbd == VK_LEFT)pk = PKLE;
			else if(kbd == VK_RIGHT)pk = PKRI;
			else if(kbd == VK_NEXT)pk = PKPD;
			else if(kbd == VK_PRIOR)pk = PKPU;
			else if(kbd == VK_INSERT)pk = PKIN;
			else if(kbd == VK_HOME)pk = PKHM;
			else if(kbd == VK_END)pk = PKED;
			else if(kbd == VK_DELETE)pk = PKDL;
			else if(kbd == VK_BACK)pk = PKBK;
			else if(kbd == VK_CANCEL || kbd == VK_ESCAPE )pk = PESC;
			else if(kbd == VK_TAB)pk = PTAB;
			else if(kbd == VK_SHIFT || kbd == VK_LSHIFT || kbd == VK_RSHIFT)pk = PSHIFT;
			else if(kbd == VK_LCONTROL || kbd == VK_RCONTROL || kbd == VK_CONTROL)pk = PCONTROL;
			else if(kbd == VK_LWIN || kbd == VK_RWIN ) pk = PSUPER;
			else if(kbd == VK_LMENU || kbd == VK_RMENU)pk = PALT;

			if(pk != PKAN){
				onskey(fwplat, pk, (nMsg == WM_KEYDOWN));
			}
			break;

		case WM_CHAR:
			kbd = (int)wParam;
			lParam = lParam & 31;
				onkey(fwplat, kbd);
			break;
		default :
			break;
		}
    return DefWindowProcW(hwnd, nMsg, wParam, lParam);
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** COMMON CONTROLS FUNCTIONS
 ** MS CONTROLS
 **
 ** common control library
 ** http://msdn.microsoft.com/en-us/library/bb773169%28v=VS.85%29.aspx
 **
 ** Toolbar Standard Button Image Index Values
 ** http://msdn.microsoft.com/en-us/library/bb760433%28VS.85%29.aspx
 **
 ** common dialogs
 ** http://msdn.microsoft.com/en-us/library/ms645524%28VS.85%29.aspx
 ** http://msdn.microsoft.com/en-us/library/aa511274.aspx
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
IPLAT i_mouse_cursor = MOUSE_CURSOR_DEFAULT;
void set_mouse_cursor(IPLAT mouse_cursor){
	HCURSOR cs = cursors[0];
	if(mouse_cursor == MOUSE_CURSOR_IBEAM) cs = cursors[1];
	else if(mouse_cursor == MOUSE_CURSOR_WAIT) cs = cursors[2];

	i_mouse_cursor = mouse_cursor;
	SetCursor(cs);
}
IPLAT get_mouse_cursor(void){
	return i_mouse_cursor;
}


#define MAX_FILE_SIZE (1024 * 64)
static char __ofn_file_ext[1024];
void on_toolbar_item_clicked(CTL_TOOLBAR_ITEM *i){
	OPENFILENAME ofn;
	char fname[MAX_FILE_SIZE];
	int id, n, l;

	if(i){
		fwplat->redraw_reason = REDRAW_MENU_ACTION;
		i->response = 0;
		if(i->iplat_toolbar == TOOLBAR_ITEM_OPENFILE || i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVEAS){
			memset(&ofn, 0, sizeof(OPENFILENAME));
			ofn.lStructSize = sizeof(OPENFILENAME);
			ofn.hwndOwner = hwndMain;
			ofn.hInstance = hinstMain;

			//file extension filter
			l = strlen(i->filefilter);
			n = 0;
			if(!l){
				sprintf(__ofn_file_ext, "*.*\0\0");
			} else {
				while(n < l){
					if(i->filefilter[n] == ';') __ofn_file_ext[n] = 0;
					else __ofn_file_ext[n] = i->filefilter[n];
					n++;
				};
				__ofn_file_ext[n] = 0; n++;
				__ofn_file_ext[n] = 0;
			}
			ofn.lpstrFilter = __ofn_file_ext;


			ofn.Flags = OFN_LONGNAMES;
			ofn.nMaxFile = MAX_FILE_SIZE;
			ofn.lpstrFile = fname;
			if(i->iplat_toolbar == TOOLBAR_ITEM_OPENFILE){
				ofn.Flags = ofn.Flags /*| OFN_FILEMUSTEXIST*/ | OFN_HIDEREADONLY;
				if(dllGetOpenFileName(&ofn)){
					i->response = fname;
					ontoolbar(fwplat, i);
					fwplat->redraw_reason = REDRAW_MENU_ACTION;
				}
			} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVEAS){
				if(dllGetSaveFileName(&ofn)){
					i->response = fname;
					ontoolbar(fwplat, i);
					fwplat->redraw_reason = REDRAW_MENU_ACTION;
				}
			}

		} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCPRINT){
			PRINTDLGA prn = {0};
			prn.lStructSize = sizeof(PRINTDLGA);
			prn.hwndOwner = hwndMain;
			dllPrintDlg(&prn);
		} else if(i->iplat_toolbar == TOOLBAR_ITEM_FONTSELECTION){
			CHOOSEFONTA fnt = {0};
			fnt.lStructSize = sizeof(CHOOSEFONTA);
			fnt.hwndOwner = hwndMain;
			fnt.Flags = CF_INITTOLOGFONTSTRUCT;
			if(__impl_winSwFontSelected)
				fnt.lpLogFont = &lfSelected;
			else
				fnt.lpLogFont = &lfStock;
			if(dllChooseFont(&fnt)){
				lfSelected = *fnt.lpLogFont;
				__impl_winSwFontSelected = 1;
				i->response = 1;
			} else if(__impl_winSwFontSelected){
				__impl_winSwFontSelected = 0;
				i->response = 0;
			}

			ontoolbar(fwplat, i);
		} else if(i->iplat_toolbar == TOOLBAR_PUSHBUTTON){
			id = (int)i->internal;
			i->sw_pushed = !i->sw_pushed;
			ontoolbar(fwplat, i);
			fwplat->redraw_reason = REDRAW_MENU_ACTION;
		} else {
			ontoolbar(fwplat, i);
			fwplat->redraw_reason = REDRAW_MENU_ACTION;
		}
	}
}
void toolbar_refresh(){
	int i, id, swstat;
	CTL_TOOLBAR_ITEM *tbitem;

	if(!hwndToolBar || !IsWindowVisible(hwndToolBar)) return;
	for(i = 0; i < lstToolBarItems->cnt; i++){
		tbitem = (CTL_TOOLBAR_ITEM*)list_at(lstToolBarItems, i);
		if(tbitem){
			id = (int)tbitem->internal;
			SendMessageW(hwndToolBar, TB_HIDEBUTTON, (WPARAM)id, (LPARAM)LOWORD(!tbitem->sw_show));
			if(tbitem->sw_show){
				swstat = SendMessageW(hwndToolBar, TB_GETSTATE, (WPARAM)id, 0);
				if(id != tbitem->sw_active)
					SendMessageW(hwndToolBar, TB_ENABLEBUTTON, (WPARAM)id, (LPARAM)LOWORD(tbitem->sw_active));

				if(tbitem->iplat_toolbar == TOOLBAR_PUSHBUTTON){
					SendMessageW(hwndToolBar, TB_CHECKBUTTON, (WPARAM)id, (LPARAM)LOWORD(tbitem->sw_pushed));
				}
			}
		}
	}
}

static TBBUTTON tbButtons[1];
int toolbar_add(CTL_TOOLBAR_ITEM *i){
	long li;
	int err=0;
	if(!i) return;
	lstToolBarItems = list_add(lstToolBarItems, LIST_ELEMENT_CAST(i));
	li = lstToolBarItems->cnt - 1;
	i->internal = (void*)li;

	memset(tbButtons, 0, sizeof(tbButtons));

	if(!hilstToolBar){
		hilstToolBar = (HIMAGELIST)dllImageList_Create(16, 16, ILC_COLOR32 | ILC_MASK, 64, 0);
		SendMessageW(hwndToolBar, TB_SETIMAGELIST, (WPARAM)0, (LPARAM)hilstToolBar);
	    SendMessageW(hwndToolBar, TB_LOADIMAGES, (WPARAM)IDB_STD_SMALL_COLOR, (LPARAM)HINST_COMMCTRL);
	}

	if(i->iplat_toolbar == TOOLBAR_ITEM_OPENFILE){
		tbButtons[0].iBitmap = MAKELONG(STD_FILEOPEN, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "Open File", 0);;
		tbButtons[0].dwData = (DWORD)i;
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else	if(i->iplat_toolbar == TOOLBAR_ITEM_DOCNEW){
		tbButtons[0].iBitmap = MAKELONG(STD_FILENEW, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "New", 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVE){
		tbButtons[0].iBitmap = MAKELONG(STD_FILESAVE, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "Save", 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSAVEAS){
		tbButtons[0].iBitmap = MAKELONG(STD_FILESAVE, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "Save As", 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_ITEM_SEPARATOR){
		tbButtons[0].iBitmap = 0;
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = TBSTYLE_SEP;
		tbButtons[0].dwData = 0;
		tbButtons[0].iString = 0;
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_PUSHBUTTON){
		tbButtons[0].iBitmap = MAKELONG(STD_COPY, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE | TBSTYLE_CHECK;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC(i->text, 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_BUTTON ){
		tbButtons[0].iBitmap = MAKELONG(STD_COPY, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC(i->text, 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCSEARCH){
		tbButtons[0].iBitmap = MAKELONG(STD_FIND, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].dwData = (DWORD)i;
		tbButtons[0].iString = (INT_PTR)WC("Search", 0);
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);
	} else if(i->iplat_toolbar == TOOLBAR_ITEM_DOCPRINT){
		tbButtons[0].iBitmap = MAKELONG(STD_PRINT, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "Print", 0);;
		tbButtons[0].dwData = (DWORD)i;
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);

	} else if(i->iplat_toolbar == TOOLBAR_ITEM_FONTSELECTION){
		tbButtons[0].iBitmap = MAKELONG(STD_PROPERTIES, 0);
		tbButtons[0].idCommand = lstToolBarItems->cnt - 1;
		tbButtons[0].fsState = TBSTATE_ENABLED;
		tbButtons[0].fsStyle = BTNS_AUTOSIZE;
		tbButtons[0].iString = (INT_PTR)WC(i->text ? i->text : "Fonts", 0);
		tbButtons[0].dwData = (DWORD)i;
	    SendMessageW(hwndToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM)sizeof(TBBUTTON), 0);
	    SendMessageW(hwndToolBar, TB_ADDBUTTONSW, (WPARAM)1, (LPARAM)&tbButtons);

	}

	if(!IsWindowVisible(hwndToolBar)){
		ShowWindow(hwndToolBar, SW_SHOW);
		resize_children();
	}

	return tbButtons[0].dwData ? 1 : 0;
}

int swTrackTabCloseBtn = 0;
TRACKMOUSEEVENT mtrack;
int mouseTabBtnIn = 0, mouseTabBtnB = 0;
LRESULT CALLBACK WinProcTabCloseBtn(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam){
	PAINTSTRUCT ps;
	HDC pdc, dc;
	RECT r;
	SIZE s;
	DWORD fgBtn, bgBtn, brdBtn;
	HPEN 	cbPen = 0, cbPenOld;
	HBRUSH	cbBrush = 0, cbBrushOld = 0;
	int i, n;
	TCITEM tie;
	char *t;

	switch (nMsg) {
		case WM_MOUSEMOVE:
			if(!swTrackTabCloseBtn){
				//track the mouse for Microsoft's strange way of providing 'mouse leave' messages
				//what could be the reason behind that?
				memset(&mtrack, 0, sizeof(TRACKMOUSEEVENT));
				mtrack.cbSize = sizeof(TRACKMOUSEEVENT);
				mtrack.dwFlags = TME_LEAVE;
				mtrack.hwndTrack = hwndTabCloseBtn;
				if(!TrackMouseEvent(&mtrack))
					swTrackTabCloseBtn = 1;
			}
			mouseTabBtnIn = 1;
			RedrawWindow(hwndTabCloseBtn, 0, 0, RDW_INTERNALPAINT);
			break;

		case WM_LBUTTONUP:
			if(mouseTabBtnB){
				i = TabCtrl_GetCurSel(hwndTab);
				memset(&tie, 0, sizeof(TCITEM));
				tie.mask = TCIF_TEXT | TCIF_PARAM;
				if(!TabCtrl_GetItem(hwndTab, i, &tie)) return 0;
				t = (char*)tie.lParam;
				if(t && ontabclose(fwplat, i + 1)){
					for(n = 0; n < lstTabPages->cnt; n++){
						if((char*)list_at(lstTabPages, n) == t){
							if(TabCtrl_DeleteItem(hwndTab, i)){
								free(t);
								lstTabPages = list_put(lstTabPages, n, 0);
							} else {

							}
							break;
						}
					}
				}

				memset(&tie, 0, sizeof(TCITEM));
				tie.mask = TCIF_TEXT | TCIF_PARAM;
				TabCtrl_GetItemCount(hwndTab);
				n = TabCtrl_GetItemCount(hwndTab);
				if(n){
					if(i > n - 1) i = n - 1;
					do{
						if(TabCtrl_GetItem(hwndTab, i, &tie)){
							t = (char*)tie.lParam;
							TabCtrl_SetCurSel(hwndTab, i);
							ontabactive(fwplat, i + 1);
							__win_redraw(0);
							break;
						}
						i--;
					}while(i >= 0);

				} else {
					lstTabPages->cnt = 0;
					resize_children();
				}

			}
			mouseTabBtnIn = 1;
			mouseTabBtnB = 0;
			RedrawWindow(hwndTabCloseBtn, 0, 0, RDW_INTERNALPAINT);
			break;

		case WM_LBUTTONDOWN:
			mouseTabBtnIn = 1;
			mouseTabBtnB = 1;
			RedrawWindow(hwndTabCloseBtn, 0, 0, RDW_INTERNALPAINT);
			break;

		case WM_MOUSELEAVE:
			mouseTabBtnIn = 0;
			mouseTabBtnB = 0;
			RedrawWindow(hwndTabCloseBtn, 0, 0, RDW_INTERNALPAINT);
			swTrackTabCloseBtn = 0;
			break;

		case WM_PAINT:
			pdc = BeginPaint(hwndTabCloseBtn, &ps);
			dc = GetDC(hwndTabCloseBtn);
			if(dc){
				GetClientRect(hwndTabCloseBtn, &r);
				dllGetTextExtentPoint(dc, TABCLOSE_TEXT, strlen(TABCLOSE_TEXT), &s);
				if(mouseTabBtnIn){
					fgBtn = GetSysColor(COLOR_HIGHLIGHTTEXT);
					bgBtn = GetSysColor(COLOR_HIGHLIGHT);
				} else {
					fgBtn = GetSysColor(COLOR_BTNTEXT);
					bgBtn = GetSysColor(COLOR_BTNFACE);
				}
				if(mouseTabBtnB){
					bgBtn = GetSysColor(COLOR_BTNSHADOW);
				}
				if(mouseTabBtnIn)
					brdBtn = GetSysColor(COLOR_BTNSHADOW);
				else
					brdBtn = GetSysColor(COLOR_BTNFACE);

				cbPen = (HPEN)dllCreatePen(PS_SOLID, 0, brdBtn);
				if(cbPen) cbPenOld = (HPEN)dllSelectObject(dc, cbPen);
				cbBrush = (HBRUSH)dllCreateSolidBrush(bgBtn);
				if(cbBrush) cbBrushOld = (HBRUSH)dllSelectObject(dc, cbBrush);
				dllRectangle(dc, r.left + 2, r.top + 2, r.right - 2, r.bottom - 2);

				dllSetBkMode(dc, TRANSPARENT);
				dllSetTextColor(dc, fgBtn);
				dllTextOut(dc, (r.right - r.left - s.cx) / 2, (r.bottom - r.top -s.cy) / 2, TABCLOSE_TEXT, strlen(TABCLOSE_TEXT));

				//cleanup
				if(cbPenOld) dllSelectObject(dc, cbPenOld);
				if(cbPen) dllDeleteObject(cbPen);
				if(cbBrushOld) dllSelectObject(dc, cbBrushOld);
				if(cbBrush) dllDeleteObject(cbBrush);
				ReleaseDC(hwndTabCloseBtn, dc);
			}
			EndPaint(hwndTabCloseBtn, &ps);
			return 0L;
			break;

	}
	return DefWindowProcW(hwnd, nMsg, wParam, lParam);
}

void* tabwnd_create(char *title){
	char *t = 0;
	TCITEMW tie = {0};
	int i=0;
	if(!title || !strlen(title)) return 0;
		t = (char*)malloc(strlen(title) + 1);
		if(t){
			memset(&tie, 0, sizeof(TCITEM));
			sprintf(t, "%s", title);
			tie.mask = TCIF_TEXT | TCIF_PARAM;
			tie.iImage = 0;
			tie.pszText = t;
			tie.lParam = (LPARAM)t;
			i = TabCtrl_GetItemCount(hwndTab);
			i = TabCtrl_InsertItem(hwndTab, i, &tie);
			lstTabPages = list_put(lstTabPages, i, LIST_ELEMENT_CAST(t));

			if(!IsWindowVisible(hwndTab)){
				ShowWindow(hwndTab, SW_SHOW);
				ShowWindow(hwndTabCloseBtn, SW_SHOW);
				resize_children();
			}
			TabCtrl_SetCurSel(hwndTab, i);
		}
	return (void*)(i + 1);
}
void* tabwnd_set_title(void *id, char *title){
	TCITEM tie;
	char *t, *tt;
	int i;
	if(hwndTab && title && strlen(title)){
		memset(&tie, 0, sizeof(TCITEM));
		tie.mask = TCIF_TEXT | TCIF_PARAM;
		i = TabCtrl_GetCurSel(hwndTab);
		if(TabCtrl_GetItem(hwndTab, i, &tie)){
			tt = (char*)tie.lParam;
			if(tt && strcmp(tt, title) == 0)
				return tt;
			t = (char*)malloc(strlen(title) + 1);
			if(!t) return tt;
			sprintf(t, "%s", title);
			tie.pszText = t;
			tie.lParam = (LPARAM)t;
			if(TabCtrl_SetItem(hwndTab, i, &tie)){
				lstTabPages = list_put(lstTabPages, i, LIST_ELEMENT_CAST(t));
				if(tt) free(tt);
				return t;
			} else {
				free(t);
				return tt;
			}
		}
	}
	return 0;
}

void statusbar_refresh(){
	WCHAR TEXT[TXTOUT_MAX];
	RECT r;
	int i[256], j, z, w;
	WPARAM wpar;
	CTL_STATUSBAR_ITEM *istat;

	GetClientRect(hwndStatus, &r);
	z = lstStatusBar->cnt;
	if(z > 256) z = 256;
	if(z > 0){
	    w = (r.right - r.left) / z;
	    for(j = 0; j < z; j++) i[j] = (j + 1) * w;
	    SendMessageW(hwndStatus, SB_SIMPLE, (WPARAM)FALSE, 0);
	    SendMessageW(hwndStatus, SB_SETPARTS, (WPARAM)z, (LPARAM)&i);

	    for(j = 0; j < z; j++){
	    	wpar = MAKEWPARAM(LOWORD(j), HIWORD(0));
	    	istat = (CTL_STATUSBAR_ITEM*)list_at(lstStatusBar, j);
	    	if(istat && istat->text[0] && istat->sw_active){
		    	WC(istat->text, TEXT);
		    	SendMessageW(hwndStatus, SB_SETTEXTW, wpar, (LPARAM)TEXT);
	    	} else {
	    		SendMessageW(hwndStatus, SB_SETTEXTW, wpar, (LPARAM)0);
	    	}

	    }
	} else {
	    SendMessageW(hwndStatus, SB_SIMPLE, (WPARAM)TRUE, 0);
	}
}
void statusbar_add(CTL_STATUSBAR_ITEM *i){
	if(!i) return;

	lstStatusBar = list_add(lstStatusBar, LIST_ELEMENT_CAST(i));
	if(!IsWindowVisible(hwndStatus)){
		ShowWindow(hwndStatus, SW_SHOW);
		resize_children();
	}
	statusbar_refresh();
}

void any_window_title(HWND wnd, char *title){
	WCHAR TITLE[TXTOUT_MAX];
	if(wnd && title){
		if(IsWindowUnicode(wnd)){
			WC(title, TITLE);
			SetWindowTextW(wnd, TITLE);
		} else {
			SetWindowTextA(wnd, title);
		}
	}
}

void set_title(char *t){
	any_window_title(hwndMain, t);
}
void set_icon(char* path){
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** MESSAGE BOXES
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
int any_box(HWND parent, UINT props, char *title, char *fmt, va_list args){
	//Format the ASCII or UTF8 input
	char buf[TXTOUT_MAX];
	vsnprintf(buf, TXTOUT_MAX, fmt, args);

	//convert UTF8 to UTF16 for windows
	WCHAR TITLE[TXTOUT_MAX], MSG[TXTOUT_MAX];
	WC(title, TITLE);
	WC(buf, MSG);

	//control with unicode support
	return MessageBoxW(parent, MSG, TITLE, props);
}

void error_msg(char *fmt, ...){
	va_list args;
	va_start(args, fmt);
	any_box(0, MB_ICONERROR | MB_OK, "Error", fmt, args);
	va_end(args);
}

void error_msgbox(const char *title, char *fmt, ...){
	if(title == 0) title = "Error";

	va_list args;
	va_start(args, fmt);
	any_box(hwndMain, MB_ICONERROR | MB_OK, title, fmt, args);
	va_end(args);
}

IPLAT yesno_msgbox(const char *title, char *fmt, ...){
	int res;
	if(title == 0) title = "Please confirm";
	va_list args;
	va_start(args, fmt);
	res = any_box(hwndMain, MB_ICONQUESTION | MB_YESNO, title, fmt, args);
	va_end(args);

	return (res == IDYES ? DLG_OK : DLG_CANCEL);
}

IPLAT msgbox(const char *title, char *fmt, ...){
	if(title == 0) title = "Info";

	va_list args;
	va_start(args, fmt);
	any_box(hwndMain, MB_ICONINFORMATION | MB_OK, title, fmt, args);
	va_end(args);

	return DLG_OK;
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** DIALOG BOXES
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
struct DIALOG {
	HWND			hwndDialog;
	LIST 			*lstCtlItems, *lstAlign;
	dialog_routine	on_altered;
	IPLAT 			cb_resp;
	void			*userdata;
}DIALOG;
struct CTL_ITEM_WIN{
	IPLAT 			controltype;
	HWND 			hwnd;
	SIZE			itmsze;
	unsigned int	id;
	int				swHidden;

	WCHAR			SelectedID[TXTOUT_MAX];
}CTL_ITEM_WIN;
struct DIALOG *CurrentDialog = 0;

void dlg_runtime_average_charsize(struct CTL_ITEM_WIN *itw, int xelems, int yelems){
	HDC hdc = 0;
	int n = 0;
	TEXTMETRIC tm;
	if(itw->hwnd){
		//if(hDialogFont) SendMessage(itw->hwnd, WM_SETFONT, (WPARAM)hDialogFont, FALSE);
		hdc = GetDC(itw->hwnd);
		if(hdc){
			if(dllGetTextMetrics(hdc, &tm)){
				itw->itmsze.cx = tm.tmAveCharWidth * xelems + DialogSepHor;
				itw->itmsze.cy = tm.tmHeight * yelems + DialogSepVer;
			}
			ReleaseDC(itw->hwnd, hdc);
		}
	}
	return n;
}
void dlg_runtime_resize_control(struct CTL_ITEM_WIN *itw, char *text){
	SIZE sze;
	RECT r;
	HDC hdc = 0;

	if(text){
		memset(&sze, 0, sizeof(sze));
		hdc = GetDC(itw->hwnd);
		dllGetTextExtentPoint(hdc, text, strlen(text), &sze);
		ReleaseDC(itw->hwnd, hdc);
		SetWindowPos(itw->hwnd, HWND_TOP, 0, 0, sze.cx + DialogSepHor, sze.cy + DialogSepVer, SWP_SHOWWINDOW);
	}
	GetWindowRect(itw->hwnd, &r);
	itw->itmsze.cx = r.right - r.left;
	itw->itmsze.cy = r.bottom - r.top;
}

void dlg_runtime_setpos(struct CTL_ITEM_WIN *item, int x, int y){
	if(item && item->hwnd)
		MoveWindow(item->hwnd, x, y, item->itmsze.cx, item->itmsze.cy, 1);
}

int dlg_runtime_get_item_pos(struct DIALOG *dlg, struct CTL_ITEM_WIN *itw, RECT *ri){
	WINDOWINFO wd, wi;
	if(!dlg || !dlg->hwndDialog || !itw || !itw->hwnd || !ri) return 0;
		memset(&wd, 0, sizeof(WINDOWINFO));
		wd.cbSize = sizeof(WINDOWINFO);
		if(!GetWindowInfo(dlg->hwndDialog, &wd)) return 0;
		memset(&wi, 0, sizeof(WINDOWINFO));
		wi.cbSize = sizeof(WINDOWINFO);
		if(!GetWindowInfo(itw->hwnd, &wi)) return 0;

		ri->top = wi.rcWindow.top - wd.rcClient.top;
		ri->bottom = wi.rcWindow.bottom - wd.rcClient.top;
		ri->left = wi.rcWindow.left -  wd.rcClient.left;
		ri->right = wi.rcWindow.right -  wd.rcClient.left;

	return 1;
}
void  dlg_runtime_draw_dialog(DIALOGBOX dialog, char *title){
	LIST_INDEX i;
	struct CTL_ITEM_WIN *itw;
	RECT rw, rc;
	NONCLIENTMETRICS ncm;
	int x = 0, y = 0, nextypos = 0, dlgwidth = 0;
	struct DIALOG *dlg = (struct DIALOG*)dialog;

	if(dlg && dlg->hwndDialog && CurrentDialog == dlg){
		any_window_title(dlg->hwndDialog, title);

    	//position all the controls in the dialog
     	x = DialogSepHor;
    	y = DialogSepVer;

		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itw = (struct CTL_ITEM_WIN*)(struct CTL_ITEM *)list_at(dlg->lstCtlItems, i);
			if(itw){
				x += DialogSepHor;
				if(itw->controltype == CTL_NEWLINE || itw->controltype == CTL_NEWLINE_DRAW){
					if(itw->swHidden == 0){
						x = DialogSepHor;
						y = y + DialogSepVer + nextypos;
						nextypos = 0;
					}
				} else {
					if(itw->swHidden == 0){
						if(nextypos < itw->itmsze.cy) nextypos = itw->itmsze.cy;
						dlg_runtime_setpos(itw, x, y);
						x += itw->itmsze.cx;
						if(dlgwidth < x) dlgwidth = x;
					} else
						x -= DialogSepHor;
				}
			}
		}


		dlgwidth = dlgwidth + DialogSepHor * 4;
		y = y + DialogSepVer * 4;
    	memset(&ncm, 0, sizeof(NONCLIENTMETRICS));
    	ncm.cbSize = sizeof(NONCLIENTMETRICS);
    	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);
    	dlgwidth += ncm.iBorderWidth * 2;
    	y += ncm.iBorderWidth * 2;
    	y += ncm.iMenuHeight;
		GetWindowRect(dlg->hwndDialog, &rw);
		GetClientRect(dlg->hwndDialog, &rc);

		if(rw.bottom - rw.top > rc.bottom - rc.top)
			nextypos = (rw.bottom - rw.top) - (rc.bottom - rc.top);
		else
			nextypos = (rc.bottom - rc.top) - (rw.bottom - rw.top);
		y += nextypos;

		MoveWindow(dlg->hwndDialog, rw.left, rw.top, dlgwidth, y, 1);
    	UpdateWindow(dlg->hwndDialog);
    	ShowWindow(dlg->hwndDialog, SW_SHOWNORMAL);
    	SetFocus(dlg->hwndDialog);
	}
}

void dlg_set_response(DIALOGBOX dialog, IPLAT resp){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg) dlg->cb_resp = resp;
}
void* dlg_get_userdata(DIALOGBOX dialog){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	return (dlg ? dlg->userdata : 0);

}
void* dlg_set_userdata(DIALOGBOX dialog, void *data){
	void *odata = 0;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg){
		odata = dlg->userdata;
		dlg->userdata = data;
	}
	return odata;
}

void dlg_item_show(DIALOGCONTROL control, int sw){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)(control);
	if(itw){
		itw->swHidden = sw;
		if(itw->hwnd) ShowWindow(itw->hwnd, sw ? SW_SHOW : SW_HIDE);
	}
}
int dlg_item_shown(DIALOGCONTROL control){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)(control);
	return (itw && itw->hwnd ? IsWindowVisible(itw->hwnd) : 0);
}
void dlg_item_enable(DIALOGCONTROL control, int sw){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)(control);
	if(itw && itw->hwnd) EnableWindow(itw->hwnd, sw);
}
int dlg_item_enabled(DIALOGCONTROL control){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)(control);
	return (itw && itw->hwnd ? IsWindowEnabled(itw->hwnd) : 0);
}

IPLAT dgl_call_callback(HWND hwnd, IPLAT icode, UINT nMsg, WPARAM wParam, LPARAM lParam){
	LIST_INDEX i;
	struct CTL_ITEM_WIN *itw;

	LPNMLISTVIEW	lv;
	LVITEMW			lvItem = {0};
	int				itmIndex;

	if(CurrentDialog && CurrentDialog->on_altered){
		CurrentDialog->cb_resp = IPLAT_NULL;
		for(i = 0; i < CurrentDialog->lstCtlItems->cnt; i++){
			itw = (struct CTL_ITEM_WIN*)list_at(CurrentDialog->lstCtlItems, i);
			if(itw &&  itw->hwnd == hwnd){

				if(itw->controltype == CTL_LIST_BOX){
					itw->SelectedID[0] = (WCHAR)0;
					lv = (LPNMLISTVIEW) lParam;
					if(lv->iItem >= 0){
						lvItem.mask = LVIF_TEXT;
						lvItem.iItem = lv->iItem;
						lvItem.cchTextMax = TXTOUT_MAX - 1;
						lvItem.pszText = &itw->SelectedID;
						SendMessageW(itw->hwnd, LVM_GETITEMTEXTW, (WPARAM)lv->iItem, (LPARAM)&lvItem);
					}
				} else if(itw->controltype == CTL_DROPBOX){
					itw->SelectedID[0] = (WCHAR)0;
					itmIndex = SendMessageW(itw->hwnd, CB_GETCURSEL, (WPARAM)0, (LPARAM)0);
					SendMessageW(itw->hwnd, CB_GETLBTEXT, (WPARAM)itmIndex, (LPARAM)&itw->SelectedID);

				}

				CurrentDialog->cb_resp = CurrentDialog->on_altered(CurrentDialog, itw, icode);
				break;
			}
		}
		return CurrentDialog->cb_resp;
	}
	return DLG_CANCEL;
}


LRESULT CALLBACK DialogWndProc(HWND hwnd, UINT nMsg, WPARAM wParam, LPARAM lParam){
	int code;
	IPLAT icode = IPLAT_NULL;
	LPNMHDR	nmh;

	switch(nMsg){
	case WM_COMMAND:
		code = HIWORD(wParam);
		if(code == CBN_SELCHANGE){			//combo box selection
			dgl_call_callback((HWND)lParam, 0, nMsg, wParam, lParam);
			return 0;
		} else if(code == EN_CHANGE){
			icode = CTL_STAT_TEXT_ALTERED;
			dgl_call_callback((HWND)lParam, icode, nMsg, wParam, lParam);
			return 0;
		} else if(code == BN_CLICKED){
			icode = CTL_STAT_CLICKED;
			dgl_call_callback((HWND)lParam, icode, nMsg, wParam, lParam);
			return 0;
		} else if(icode != IPLAT_NULL){
			dgl_call_callback((HWND)lParam, icode, nMsg, wParam, lParam);
			return 0;
		}
		break;

	case WM_NOTIFY:
		nmh = (LPNMHDR)lParam;
		if(nmh){
			if(nmh->code == LVN_ITEMCHANGED)	//list view selection
				dgl_call_callback(nmh->hwndFrom, 0, nMsg, wParam, lParam);
		}
		break;

	case WM_CLOSE:
		ShowWindow(hwnd, SW_HIDE);
		return 0;
		break;
	}

	return DefWindowProcW(hwnd, nMsg, wParam, lParam);
}


IPLAT dlg_domodal(DIALOGBOX dialog, char *title, dialog_routine fct){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	IPLAT res = IPLAT_NULL;
	MSG msg;

	if(fct && dlg && dlg->hwndDialog && dlg->lstCtlItems && dlg->lstCtlItems->cnt){
		dlg->cb_resp = IPLAT_NULL;
		dlg->on_altered = fct;
		dlg->cb_resp = CTL_CB_CONTINUE;
		CurrentDialog = dlg;

		dlg_runtime_draw_dialog(dialog, title);

		while(!swWinExit && IsWindowVisible(dlg->hwndDialog)){
			if(dlg->cb_resp == CTL_CB_CLOSEDLG)
				SendMessageW(dlg->hwndDialog, WM_CLOSE, 0, FALSE);

			GetMessageW( &msg, 0, 0, 0);
			TranslateMessage(&msg);
			DispatchMessageW(&msg);
		}

		res = dlg->cb_resp;
	}


	CurrentDialog = 0;

	return res;
}

void dlg_set_focus(DIALOGBOX dialog, DIALOGCONTROL item){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)(item);
	if(itw && itw->hwnd && IsWindowVisible(itw->hwnd))
			SetFocus(itw->hwnd);
}

HWND any_control_create(char *classname, char *text, DWORD style, HWND hParent){
	static WCHAR CLASSNAME[TXTOUT_MAX], TEXT[TXTOUT_MAX];

	WC(classname, CLASSNAME);
	WC(text, TEXT);

	return
	CreateWindowExW(	0,
						CLASSNAME, text ? TEXT : NULL,
						style, CW_USEDEFAULT, CW_USEDEFAULT, 0, 0,
		     	 	 	hParent, NULL, hinstMain, NULL);

}

char* __impl_dgl_initString_number(char *initstring, int *out){
	char 	tmp[16],
			*head = tmp,
			*tail = initstring;

	while(tail && *tail){
		if(*tail == ';'){
			tail++;
			break;
		}
		else if(*tail >= '0' && *tail <= '9'){
			*head = *tail;
			head++;
		}
		tail++;
	}
	*head = 0;
	*out = atoi(tmp);

	return tail;
}

DIALOGBOX dlg_add_item(DIALOGBOX dialog, IPLAT ctlType, unsigned int id, char *initString){
	RECT r;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct CTL_ITEM_WIN *itw;
	LVCOLUMNW	lvColumn = {0};
	WCHAR	TEXT[TXTOUT_MAX];
	char	text[TXTOUT_MAX], *head, *tail;
	int 	cnt, columns = 0;
	int		wdt, hgt;

	if(!dlg){
		dlg = (struct DIALOG*) malloc(sizeof(struct DIALOG));
		if(!dlg) return 0;
		memset(dlg, 0, sizeof(struct DIALOG));
	    dlg->hwndDialog = any_control_create(WC_CUSTOMDIALOG, 0, WS_SYSMENU | WS_CLIPCHILDREN | WS_THICKFRAME, hwndMain);
	    if(!dlg->hwndDialog){
	    	free(dlg);
	    	return 0;
	    }

		dlg->lstCtlItems = list_new(LIST_MIN_BLOCK);
		dlg->lstAlign = list_new(LIST_MIN_BLOCK);
	}

	itw = (struct CTL_ITEM_WIN*)malloc(sizeof(struct CTL_ITEM_WIN));
	if(!itw) return dlg;
	memset(itw, 0, sizeof(struct CTL_ITEM_WIN));
	dlg->lstCtlItems = list_add(dlg->lstCtlItems, LIST_ELEMENT_CAST(itw));
	itw->controltype = ctlType;
	itw->id = id;

	#define PLATWIN_BORDER			2
	#define PLATWIN_BUTTON_WIDTH 	16
	#define PLATWIN_TEXT_WDT		(PLATWIN_BUTTON_WIDTH * 2)
	#define PLATWIN_MULTI_HGT		6

	if(!itw->hwnd){
		switch(ctlType){
		case CTL_LABEL:				/*initstring label text*/
		    itw->hwnd = any_control_create(WC_STATIC, initString, WS_VISIBLE | WS_CHILD | BS_LEFTTEXT, dlg->hwndDialog);
			dlg_runtime_average_charsize(itw, initString ? strlen(initString) : PLATWIN_TEXT_WDT + PLATWIN_BORDER, 1);
			break;

		case CTL_BUTTON:			/*initstring button text*/
		    itw->hwnd = any_control_create(WC_BUTTON, initString,  WS_TABSTOP | WS_VISIBLE | WS_CHILD | BS_PUSHBUTTON, dlg->hwndDialog);
			dlg_runtime_average_charsize(itw, initString ? strlen(initString) + PLATWIN_BORDER : PLATWIN_BUTTON_WIDTH + PLATWIN_BORDER, 1);
			break;

		case CTL_CHECK_BUTTON:		/*initstring check box text*/
		    itw->hwnd = any_control_create(WC_BUTTON, initString,  WS_TABSTOP | WS_VISIBLE | WS_CHILD | BS_LEFTTEXT | BS_CHECKBOX, dlg->hwndDialog);
			dlg_runtime_average_charsize(itw, initString ? strlen(initString) + PLATWIN_BORDER : PLATWIN_BORDER, 1);
			break;

		case CTL_PUSH_BUTTON:		/*initstring button text*/
		    itw->hwnd = any_control_create(WC_BUTTON, initString, WS_TABSTOP | WS_VISIBLE | WS_CHILD | BS_FLAT, dlg->hwndDialog);
			dlg_runtime_average_charsize(itw, initString ? strlen(initString) + PLATWIN_BORDER : PLATWIN_BUTTON_WIDTH + PLATWIN_BORDER, 1);
			break;

		case CTL_TEXT:				/*initstring x; where x = width in characters*/
		    itw->hwnd = any_control_create(WC_EDIT, 0, WS_TABSTOP | WS_VISIBLE | WS_CHILD | WS_BORDER | ES_NOHIDESEL | ES_AUTOHSCROLL, dlg->hwndDialog);

		    initString = __impl_dgl_initString_number(initString, &wdt);
			if(wdt == 0) wdt = PLATWIN_TEXT_WDT;

		    dlg_runtime_average_charsize(itw, wdt + PLATWIN_BORDER, 1);
			break;

		case CTL_TEXT_MULTI_LINE:	/*initstring x;y; where x = width in characters and y = amount of lines*/
		    itw->hwnd = any_control_create(WC_EDIT, 0, WS_TABSTOP | WS_VISIBLE | WS_CHILD | WS_BORDER | WS_VSCROLL | WS_HSCROLL | ES_MULTILINE | ES_AUTOHSCROLL | ES_AUTOVSCROLL | BS_LEFTTEXT , dlg->hwndDialog);

		    initString = __impl_dgl_initString_number(initString, &wdt);
			if(wdt == 0) wdt = PLATWIN_TEXT_WDT;

		    initString = __impl_dgl_initString_number(initString, &hgt);
			if(hgt == 0) wdt = PLATWIN_MULTI_HGT;

		    dlg_runtime_average_charsize(itw, wdt + PLATWIN_BORDER, hgt);
		    break;

		case CTL_DROPBOX:			/*initstring x; where x = width in characters*/
		    itw->hwnd = any_control_create(WC_COMBOBOX, 0, WS_TABSTOP | WS_VISIBLE | WS_CHILD | WS_OVERLAPPED | CBS_DROPDOWN | CBS_HASSTRINGS | CBS_SIMPLE, dlg->hwndDialog);

		    initString = __impl_dgl_initString_number(initString, &wdt);
			if(wdt == 0) wdt = PLATWIN_TEXT_WDT;

		    dlg_runtime_average_charsize(itw, wdt + PLATWIN_BORDER, 1);
			break;

		case CTL_LIST_BOX:			/*initstring title "x;y;column 1;title column 2;...;" where x = width in characters and y = height in lines of the box*/
		    itw->hwnd = any_control_create(WC_LISTVIEW, 0, WS_TABSTOP | WS_VISIBLE | WS_CHILD | WS_BORDER | LBS_NOTIFY | LVS_REPORT, dlg->hwndDialog);

		    initString = __impl_dgl_initString_number(initString, &wdt);
			if(wdt == 0) wdt = PLATWIN_TEXT_WDT;

		    initString = __impl_dgl_initString_number(initString, &hgt);
			if(hgt == 0) wdt = PLATWIN_MULTI_HGT;


		    dlg_runtime_average_charsize(itw, wdt + PLATWIN_BORDER, hgt);

			lvColumn.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT ;
			lvColumn.cx = 0;
			lvColumn.fmt = LVCFMT_RIGHT;	//LVCFMT_FIXED_WIDTH and HDF_FIXEDWIDTH are not defined (value of 0x100 ??) but it doesn't work anyway
			lvColumn.pszText = &TEXT;

			//create the 1st hidden column for ID's
			WC("ID", &TEXT);
			SendMessageW(itw->hwnd, LVM_INSERTCOLUMNW, (WPARAM)columns, (LPARAM)&lvColumn);
			columns++;

			//create the columns from the init string
			head = tail = initString;
			while(tail && *tail){
				if(*tail == ';'){
					cnt = tail - head;
					memcpy(text, head, cnt);
					text[cnt] = 0;
					WC(text, TEXT);
					lvColumn.cx = cnt * 16;
					SendMessageW(itw->hwnd, LVM_INSERTCOLUMNW, (WPARAM)columns, (LPARAM)&lvColumn);
					columns++;
					head = tail + 1;
				}
				tail++;
			}
			SendMessageW(itw->hwnd, LVM_SETEXTENDEDLISTVIEWSTYLE, (WPARAM)0, (LPARAM)LVS_EX_FULLROWSELECT);


			break;

		case CTL_NEWLINE_DRAW:
			break;

		case CTL_NEWLINE:
			break;

		default:
			break;
		}

	}

	return (void*)dlg;
}

DIALOGCONTROL dlg_item_find(DIALOGBOX dialog, unsigned int id){
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct CTL_ITEM_WIN *itw = 0;
	LIST_INDEX i;
	if(id && dlg && dlg->lstCtlItems){
		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itw = (struct CTL_ITEM_WIN*)list_at(dlg->lstCtlItems, i);
			if(itw && itw->id == id)
				return itw;
		}
	}
	return itw;
}

DIALOGCONTROL dlg_item_next(DIALOGBOX dialog, DIALOGCONTROL lastitem){
	int sw = 0;
	LIST_INDEX i;
	struct CTL_ITEM_WIN *itw;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	if(dlg){
		for(i = 0; i < dlg->lstCtlItems->cnt; i++){
			itw = (struct CTL_ITEM_WIN*)list_at(dlg->lstCtlItems, i);
			if(!itw) continue;
			if(!lastitem || sw) return itw;
			if(itw == lastitem) sw = 1;
		}
	}
	return 0;
}
void dlg_item_set_value(DIALOGCONTROL control, char *value){
	struct CTL_ITEM_WIN *itm = (struct CTL_ITEM_WIN*)control;
	WCHAR	TEXT[TXTOUT_MAX];
	char	text[TXTOUT_MAX], *head, *tail;
	int		n;

	if(itm && itm->hwnd){
		if(itm->controltype == CTL_DROPBOX){
			SendMessageW(itm->hwnd, CB_RESETCONTENT, 0, 0);
			head = value;
			while(head && *head){
				tail = strchr(head, '\n');
				if(!tail) tail = value + strlen(value);
				n = tail - head;
				memcpy(text, head, n);
				text[n + 1] = 0;
				WC(text, TEXT);
				SendMessageW(itm->hwnd, CB_ADDSTRING, 0, TEXT);
				head = tail + 1;
			}
			SendMessageW(itm->hwnd, CB_SETCURSEL, (WPARAM)0, (LPARAM)0);
			return;
		}

		if(itm->controltype == CTL_TEXT){
			any_window_title(itm->hwnd, value);
			return;
		}

		if(itm->controltype == CTL_TEXT_MULTI_LINE){
			//replace \n by \r\n
			text[0] = 0;
			head = text;
			tail = value;
			while(tail && *tail){
				if(*tail == '\n'){
					*head = '\r'; head++;
					*head = '\n'; head++;
					tail++;
					continue;
				}

				if(*tail == '\r'){
					tail++;
					continue;
				}

				*head = *tail;
				head++;
				tail++;
			}

			*head = 0;

			any_window_title(itm->hwnd, text);
			return;
		}

	}

	/*value format is:
	 * id;value column 1\nvalue column 2\n...\n
	 * id;value column 1\nvalue column 2\n...\n
	 * ...
	 *
	 * https://stackoverflow.com/questions/8966330/creating-a-table-using-win32-api
	 */
	if(itm->controltype == CTL_LIST_BOX){
		int cnt, items = 0, insertedItems = 0;
		LVITEMW		lvItem;

		SendMessageW(itm->hwnd, LVM_DELETEALLITEMS, (WPARAM)0, (LPARAM)0);

		//item set
		memset(&lvItem, 0, sizeof(LVITEMW));
		lvItem.mask = LVIF_TEXT;
		lvItem.pszText = &TEXT;


		head = tail = value;
		while(tail && *tail){
			if(*tail == ';'){
				//add a line in column, the line = lvItem
				cnt = tail - head;
				memcpy(text, head, cnt);
				text[cnt] = 0;
				WC(text, TEXT);
				lvItem.iItem = insertedItems + 1;	//append the line
				lvItem.iSubItem = 0;
				insertedItems = SendMessageW(itm->hwnd, LVM_INSERTITEMW, (WPARAM)0, (LPARAM)&lvItem);
				items = 1;

				head = tail + 1;
			} else if(*tail == '\n'){
				//feed the next columns of the line
				cnt = tail - head;
				memcpy(text, head, cnt);
				text[cnt] = 0;
				WC(text, TEXT);
				lvItem.iSubItem = items;
				SendMessageW(itm->hwnd, LVM_SETITEMTEXTW, (WPARAM)(insertedItems), (LPARAM)&lvItem);

				head = tail + 1;
				items++;
			}

			tail++;
		}

		return;
	}


}

char	__impl_dlg_item_get_value_text[TXTOUT_MAX];
char* dlg_item_get_value(DIALOGCONTROL control){
	char 	*head, *tail;
	WCHAR	TEXT[TXTOUT_MAX];
	int 	len;

	struct CTL_ITEM_WIN *itm = (struct CTL_ITEM_WIN*)control;
	if(itm && itm->hwnd){
		switch(itm->controltype){
		case CTL_BUTTON:
			break;

		case CTL_CHECK_BUTTON:
			break;

		case CTL_PUSH_BUTTON:
			break;

		case CTL_TEXT:
			len = GetWindowTextW(itm->hwnd, &TEXT, TXTOUT_MAX - 1);
			TEXT[len] = (WCHAR)0;
			return UTF8(&TEXT, 0);
		    break;

		case CTL_TEXT_MULTI_LINE:
			len = GetWindowTextW(itm->hwnd, &TEXT, TXTOUT_MAX - 1);
			TEXT[len] = (WCHAR)0;
			head = &__impl_dlg_item_get_value_text;
			tail = UTF8(&TEXT, 0);
			while(*tail){
				if(*tail != '\r'){
					*head = *tail;
					head++;
				}
				tail++;
			}
			*head = 0;
			return &__impl_dlg_item_get_value_text;
			break;

		case CTL_DROPBOX:
			return UTF8(itm->SelectedID, 0);
			break;

		case CTL_LIST_BOX:
			return UTF8(itm->SelectedID, 0);
			break;

		default:
			break;
		}
	}
	return 0;
}
void dlg_item_reset(DIALOGCONTROL control){
}
char* dlg_item_get_id(DIALOGCONTROL control){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)control;
	return (itw ? itw->id : 0);
}
int dlg_item_get_value_size(DIALOGCONTROL control){
	return 0;
}
IPLAT dlg_item_type(DIALOGCONTROL control){
	struct CTL_ITEM_WIN *itw = (struct CTL_ITEM_WIN*)control;
	return (itw ? itw->controltype : 0);
}
DIALOGBOX dlg_destroy(DIALOGBOX dialog){
	struct CTL_ITEM_WIN *itw;
	struct DIALOG *dlg = (struct DIALOG*)dialog;
	struct DIALOG_ALIGN *da;
	if(dlg){
		while(dlg->lstCtlItems && dlg->lstCtlItems->cnt){
			itw = (struct CTL_ITEM_WIN*) list_pop(dlg->lstCtlItems);
			if(itw){
				if(itw->hwnd) DestroyWindow(itw->hwnd);
				free(itw);
			}
		}
		dlg->lstCtlItems = list_destroy(dlg->lstCtlItems);

		while(dlg->lstAlign && dlg->lstAlign->cnt){
			da = (struct DIALOG_ALIGN*) list_pop(dlg->lstAlign);
			if(da) free(da);
		}
		dlg->lstAlign = list_destroy(dlg->lstAlign);

		if(dlg->hwndDialog) DestroyWindow(dlg->hwndDialog);
		free(dlg);
	}
	return 0;
}


char* get_user(void){
	return getenv("USERNAME");
}
char* get_home(void){
	return getenv("HOMEPATH");
}

char get_path_seperator(void){
	return '\\';
}
int makedir(char *path){
	return mkdir(path);
}
int ls_exists(char *path){
	struct stat	 st;
	return (path && (0 == stat(path, &st)) ? 1 : 0);
}

/****************************************************************************************************************************************************
 ****************************************************************************************************************************************************
 ** WINMAIN
 ****************************************************************************************************************************************************
 ****************************************************************************************************************************************************/
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow){
	char *h, *t, *arg, *a;
	int posx = CW_USEDEFAULT, posy = CW_USEDEFAULT, wdt = CW_USEDEFAULT, hgt = CW_USEDEFAULT, i;
	MSG msg;
	LIST_INDEX j;

	WC("WinApp", szMainWndClass);
	WC(TOOLBARCLASSNAME, szToolbarWndClass);
	WC(WC_TABCONTROL, szTabControlWndClass);
	WC("ChildClass", szChildWndClass);
	WC("tabClassButton", szTabButtonClass);
	WC(STATUSCLASSNAME, szTabStatusBarClass);
	WC(WC_CUSTOMDIALOG, szDialogWindowClass);


	//todo: DirectD2 measures wrong fonts, window rect with high dpi (gdi and gdiplus are ok)
	SetProcessDPIAware();

	hinstMain = hInstance;
	fwplat = (PLATFORM_APP*)malloc(sizeof(PLATFORM_APP));
	if(fwplat){
		memset(fwplat, 0, sizeof(PLATFORM_APP));
	} else
		return 1;

	platform_app_init(fwplat);

	fwplat->sys.lstArgv = list_new(16);
	while(lpCmdLine){
		h = lpCmdLine;
		while(*h && *h == ' ') h++;
		t = h;
		while(*t && *t != ' ') t++;
		i = t - h + 1;
		if(i > 1){
			arg = malloc(i);
			if(!arg) break;
			memset(arg, 0, i);
			a = arg;
			while(h < t){
				if(*h == '\"'){	//exclude the quotes of arguments containing spaces
					h++;
					continue;
				}
				*a++ = *h++;
			}
			fwplat->sys.lstArgv = list_add(fwplat->sys.lstArgv, LIST_ELEMENT_CAST(arg));
		}
		if(!*t) break;
		lpCmdLine = t;
	}

	fwplat->gui.title = set_title;
	fwplat->gui.icon = set_icon;

	fwplat->gui.line = line;
	fwplat->gui.rectangle = winrectangle;
	fwplat->gui.circle = wincircle;
	fwplat->gui.polygon = winpolygon;
	fwplat->gui.linepath = winlinepath;
	fwplat->gui.dashline = dashline;
	fwplat->gui.putpixel = putpixel;

	fwplat->gui.set_intensity = set_intensity;
	fwplat->gui.set_alpha = set_alpha;
	fwplat->gui.set_fcolor = __impl_set_fcolorn;
	fwplat->gui.set_bcolor = __impl_set_bcolorn;
	fwplat->gui.set_sys_color = set_sys_color;
	fwplat->gui.swap_colors = __imp_swap_colors;

	fwplat->gui.maxx = maxx;
	fwplat->gui.maxy = maxy;
	fwplat->gui.resolutionx = resolutionx;
	fwplat->gui.resolutiony = resolutiony;

	fwplat->gui.textmode = textmode;
	fwplat->gui.textout = textout;
	fwplat->gui.lenpixels = lenpixels;

	fwplat->gui.caret_set = caret_set;
	fwplat->gui.clrscr = clrscr;
	fwplat->gui.fontwdt = fontwdt;
	fwplat->gui.fonthgt = fonthgt;
	fwplat->gui.redraw = __win_redraw;
	fwplat->gui.screencopy = screencopy;
	fwplat->gui.screenput = screenput;
	fwplat->gui.screendelete = screendelete;
	fwplat->gui.font_use = font_use;
	fwplat->gui.font_face = font_face;

	fwplat->ctl.set_mouse_cursor = set_mouse_cursor;
	fwplat->ctl.get_mouse_cursor = get_mouse_cursor;
	fwplat->ctl.error_msg = error_msg;
	fwplat->ctl.sbar_set = sbar_set;
	fwplat->ctl.sbar_show = sbar_show;
	fwplat->ctl.tabwnd_create = tabwnd_create;
	fwplat->ctl.tabwnd_set_title = tabwnd_set_title;
	fwplat->ctl.toolbar_add = toolbar_add;
	fwplat->ctl.toolbar_refresh = toolbar_refresh;
	fwplat->ctl.statusbar_add = statusbar_add;
	fwplat->ctl.statusbar_refresh = statusbar_refresh;
	fwplat->ctl.error_msg = error_msgbox;
	fwplat->ctl.yesno_msg = yesno_msgbox;
	fwplat->ctl.msgbox = msgbox;

	fwplat->ctl.dlg_add_item = dlg_add_item;
	fwplat->ctl.dlg_item_find = dlg_item_find;
	fwplat->ctl.dlg_item_next = dlg_item_next;
	fwplat->ctl.dlg_set_focus = dlg_set_focus;
	fwplat->ctl.dlg_update = dlg_runtime_draw_dialog;
	fwplat->ctl.dlg_set_response = dlg_set_response;
	fwplat->ctl.dlg_domodal = dlg_domodal;
	fwplat->ctl.dlg_destroy = dlg_destroy;

	fwplat->ctl.dlg_item_show = dlg_item_show;
	fwplat->ctl.dlg_item_shown = dlg_item_shown;
	fwplat->ctl.dlg_item_enable = dlg_item_enable;
	fwplat->ctl.dlg_item_enabled = dlg_item_enabled;
	fwplat->ctl.dlg_item_type = dlg_item_type;
	fwplat->ctl.dlg_item_set_value = dlg_item_set_value;
	fwplat->ctl.dlg_item_get_value = dlg_item_get_value;
	fwplat->ctl.dlg_item_reset = dlg_item_reset;
	fwplat->ctl.dlg_item_get_id = dlg_item_get_id;
	fwplat->ctl.dlg_item_get_value_size = dlg_item_get_value_size;
	fwplat->ctl.dlg_get_userdata = dlg_get_userdata;
	fwplat->ctl.dlg_set_userdata = dlg_set_userdata;	//attaches user usable data to the dialog, returns previous user data


	fwplat->sys.doevents = doevents;
	fwplat->sys.exit = win_impl_exit;
	fwplat->sys.total_memory = total_memory;
	fwplat->sys.timed_call = timed_call;
	fwplat->sys.time = __impl_time;
	fwplat->sys.time_delta = __impl_time_delta;
	fwplat->sys.get_home = get_home;
	fwplat->sys.get_user = get_user;
	fwplat->sys.get_path_seperator = get_path_seperator;
	fwplat->sys.mkdir = makedir;
	fwplat->sys.ls_exists = ls_exists;

	wdt = GetSystemMetrics(SM_CXVIRTUALSCREEN);
	hgt = GetSystemMetrics(SM_CYVIRTUALSCREEN);
	if(fwplat->wShapeHint == WS_SQUARE){
		wdt = MIN(wdt, hgt);
		hgt = wdt;
	}
	wdt *= DEFAULT_WINDOW_SIZE;
	hgt *= DEFAULT_WINDOW_SIZE;

	#ifdef PLATFORM_GDI_PLUS
		gdpInput.GdiplusVersion = 1;
		gdpInput.DebugEventCallback = NULL;
		gdpInput.SuppressBackgroundThread = FALSE;
		gdpInput.SuppressExternalCodecs = TRUE;
		GpStatus gdiStat = GdiplusStartup(&gdpToken, &gdpInput, NULL);
		if(gdiStat != Ok){
			error_msg("GDI PLUS startup failure. Status: %d.", gdiStat);
			return 1;
		}
	#endif

	#ifdef PLATFORM_DIRECT2D
		hDx2d = LoadLibrary(DX2D_DLL_NAME);
		if(!hDx2d){
			error_msg( DX2D_DLL_NAME " not found");
			return 1;
		}

		dllDX2D_init = GetProcAddress(hDx2d, "dx2d_init");
		dllDX2D_exit = GetProcAddress(hDx2d, "dx2d_exit");
		dllDX2D_render_start = GetProcAddress(hDx2d, "dx2d_render_start");
		dllDX2D_render_end = GetProcAddress(hDx2d, "dx2d_render_end");
		dllDX2D_color_pen = GetProcAddress(hDx2d, "dx2d_color_pen");
		dllDX2D_color_brush = GetProcAddress(hDx2d, "dx2d_color_brush");
		dllDX2D_fontw = GetProcAddress(hDx2d, "dx2d_fontw");
		dllDX2D_fonth = GetProcAddress(hDx2d, "dx2d_fonth");
		dllDX2D_usefont = GetProcAddress(hDx2d, "dx2d_usefont");
		dllDX2D_set_text = GetProcAddress(hDx2d, "dx2d_set_text");
		dllDX2D_lentextw = GetProcAddress(hDx2d, "dx2d_lentextw");
		dllDX2D_outtextw = GetProcAddress(hDx2d, "dx2d_outtextw");
		dllDX2D_cls = GetProcAddress(hDx2d, "dx2d_cls");
		dllDX2D_line = GetProcAddress(hDx2d, "dx2d_line");
		dllDX2D_dashline = GetProcAddress(hDx2d, "dx2d_dashline");
		dllDX2D_rectangle = GetProcAddress(hDx2d, "dx2d_rectangle");
		dllDX2D_circle = GetProcAddress(hDx2d, "dx2d_circle");
		dllDX2D_polygon = GetProcAddress(hDx2d, "dx2d_polygon");
		dllDX2D_linepath = GetProcAddress(hDx2d, "dx2d_linepath");

		if(	!dllDX2D_init || !dllDX2D_exit ||
			!dllDX2D_render_start || !dllDX2D_render_end ||
			!dllDX2D_color_pen || !dllDX2D_color_brush ||
			!dllDX2D_fontw || !dllDX2D_fonth ||
			!dllDX2D_usefont || !dllDX2D_set_text ||
			!dllDX2D_lentextw || !dllDX2D_outtextw ||
			!dllDX2D_cls ||
			!dllDX2D_line || !dllDX2D_dashline ||
			!dllDX2D_rectangle || !dllDX2D_circle ||
			!dllDX2D_polygon || !dllDX2D_linepath
		){
		   	FreeLibrary(hDx2d);
			error_msg( DX2D_DLL_NAME " routines were not found");
			return 1;
		}

		dllDX2D_iface.WC = &WC;
		dxResult = dllDX2D_init(&dllDX2D_iface, fwplat->sw_double_buffer);
		if(!SUCCEEDED(dxResult)){
			error_msg("DirectX2D startup failure. Status: 0x%X.", dxResult);
			return 1;
		}
	#endif


	hWinGdi = LoadLibrary("gdi32.dll");
	hWinCtrl = LoadLibrary("Comctl32.dll");
	hWinDlg  = LoadLibrary("Comdlg32.dll");
	hUser32 = LoadLibrary("User32.dll");
	hKernel32 = LoadLibrary("Kernel32.dll");

	if(hKernel32)
		dllGlobalMemoryStatusEx = GetProcAddress(hKernel32, "GlobalMemoryStatusEx");

	if(hWinGdi && hWinCtrl && hWinDlg && hUser32){
		/*GDI+ routines*/
		dllLineTo = GetProcAddress(hWinGdi, "LineTo");
		dllMoveToEx = GetProcAddress(hWinGdi, "MoveToEx");
		dllRectangle = GetProcAddress(hWinGdi, "Rectangle");
		dllFillRect = GetProcAddress(hUser32, "FillRect");
		dllPolygon = GetProcAddress(hWinGdi, "Polygon");
		dllPolyline = GetProcAddress(hWinGdi, "Polyline");
		dllEllipse = GetProcAddress(hWinGdi, "Ellipse");
		dllArc = GetProcAddress(hWinGdi, "Arc");
		dllTextOut = GetProcAddress(hWinGdi, "TextOutA");
		dllTextOutW = GetProcAddress(hWinGdi, "TextOutW");
		dllGetTextExtentPoint = GetProcAddress(hWinGdi, "GetTextExtentPoint32A");
		dllGetTextExtentPointW = GetProcAddress(hWinGdi, "GetTextExtentPoint32W");
		dllSetPixel = GetProcAddress(hWinGdi, "SetPixel");
		dllCreatePen = GetProcAddress(hWinGdi, "CreatePen");
		dllCreateSolidBrush = GetProcAddress(hWinGdi, "CreateSolidBrush");
		dllSelectObject = GetProcAddress(hWinGdi, "SelectObject");
		dllDeleteObject = GetProcAddress(hWinGdi, "DeleteObject");
		dllSetTextColor = GetProcAddress(hWinGdi, "SetTextColor");
		dllSetBkColor = GetProcAddress(hWinGdi, "SetBkColor");
		dllGetTextFace = GetProcAddress(hWinGdi, "GetTextFaceA");
		dllCreateFont  = GetProcAddress(hWinGdi, "CreateFontA");
		dllCreateCompatibleDC = GetProcAddress(hWinGdi, "CreateCompatibleDC");
		dllBitBlt = GetProcAddress(hWinGdi, "BitBlt");
		dllCreateCompatibleBitmap = GetProcAddress(hWinGdi, "CreateCompatibleBitmap");
		dllGetTextMetrics = GetProcAddress(hWinGdi, "GetTextMetricsA");
		dllSetBkMode = GetProcAddress(hWinGdi, "SetBkMode");
		if(	!dllGetTextExtentPointW || !dllTextOutW || !dllTextOut || !dllGetTextExtentPoint ||
			!dllLineTo || !dllMoveToEx || !dllRectangle || !dllFillRect || !dllPolygon || !dllPolyline || !dllEllipse || !dllArc || !dllSetPixel || !dllCreatePen || !dllCreateSolidBrush ||
			!dllSelectObject || !dllDeleteObject || !dllSetTextColor || !dllSetBkColor || !dllGetTextFace || !dllCreateFont ||
			!dllCreateCompatibleDC || !dllBitBlt || !dllCreateCompatibleBitmap || !dllGetTextMetrics ||
			!dllSetBkMode
		){
			error_msg("gdi32.dll failure");
			return 1;
		}

		/*common controls*/
		#define COMCTL_MINIMAL_INIT_VISUALS 0x0300
		dllInitCommonControlsEx = GetProcAddress(hWinCtrl, "InitCommonControlsEx");
		dllImageList_Create = GetProcAddress(hWinCtrl, "ImageList_Create");
		dllImageList_Destroy = GetProcAddress(hWinCtrl, "ImageList_Destroy");
		if(	!dllInitCommonControlsEx ||
			!dllImageList_Create || !dllImageList_Destroy
		){
			error_msg("Comctl32.dll or User32.Dll failure");
			return 1;
		}
		memset(&initCtrl, 0,sizeof(INITCOMMONCONTROLSEX));
		initCtrl.dwSize = sizeof(INITCOMMONCONTROLSEX);
		initCtrl.dwICC = COMCTL_MINIMAL_INIT_VISUALS | ICC_BAR_CLASSES | ICC_TAB_CLASSES;
		if(!dllInitCommonControlsEx(&initCtrl)){
			error_msg("Common Controls failure");
			return 1;
		}

		/*Common dialog routines*/
		dllGetOpenFileName = GetProcAddress(hWinDlg, "GetOpenFileNameA");
		dllGetSaveFileName = GetProcAddress(hWinDlg, "GetSaveFileNameA");
		dllPrintDlg = GetProcAddress(hWinDlg, "PrintDlgA");
		dllChooseFont = GetProcAddress(hWinDlg, "ChooseFontA");
		if(	!dllGetOpenFileName || !dllGetSaveFileName || !dllPrintDlg || !dllChooseFont){
			error_msg("Comdlg32.dll failure");
			return 1;
		}

		/*the main parent window*/
	    HICON hico = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(200));
	    memset(&wndclass, 0, sizeof(WNDCLASSEXW));
	    wndclass.lpszClassName = szMainWndClass;
	    wndclass.cbSize = sizeof(WNDCLASSEXW);
	    wndclass.style = CS_HREDRAW | CS_VREDRAW;
	    wndclass.lpfnWndProc = MainWndProc;
	    wndclass.hbrBackground = 0;
	    wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	    wndclass.hIcon = hico;
	    wndclass.hIconSm = hico;
	    RegisterClassExW(&wndclass);
	    hwndMain = CreateWindowExW(0, szMainWndClass, NULL, WS_OVERLAPPEDWINDOW,
	    		     posx, posy, wdt, hgt,
	    		     NULL, NULL, hInstance, NULL);

	    if(hwndMain){
	    	/*Toolbar window*/
	    	lstToolBarItems = list_new(16);
	    	hwndToolBar = CreateWindowExW(0, szToolbarWndClass, NULL,
											WS_CHILD | TBSTYLE_FLAT | CCS_NODIVIDER | TBSTYLE_TOOLTIPS,
											0, 0, 0, 0,
											hwndMain, NULL, hInstance, NULL);

	        /*The Tab window*/
	    	lstTabPages = list_new(16);
	    	NONCLIENTMETRICS ncm;
	    	memset(&ncm, 0, sizeof(NONCLIENTMETRICS));
	    	ncm.cbSize = sizeof(NONCLIENTMETRICS);
	    	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);
	    	TabHeight = ncm.iMenuHeight;
			hwndTab = CreateWindowExW(0, szTabControlWndClass, NULL,
									WS_CHILD | WS_CLIPSIBLINGS | TCS_TABS | TCS_FOCUSNEVER,
									0, 0, 0, TabHeight,
									hwndMain, NULL, hInstance, NULL);


	    	//hTabFont = CreateFontIndirect((const LOGFONT*)&ncm.lfMenuFont);
			//SendMessage(hwndTab, WM_SETFONT, (WPARAM)hTabFont, FALSE);

			/*The Tab Close Button*/
		    memset(&tabclassbutton, 0, sizeof(WNDCLASSEXW));
		    tabclassbutton.lpszClassName = szTabButtonClass;
		    tabclassbutton.cbSize = sizeof(WNDCLASSEXW);
		    tabclassbutton.style = CS_HREDRAW | CS_VREDRAW;
		    tabclassbutton.lpfnWndProc = WinProcTabCloseBtn;
		    tabclassbutton.hbrBackground = (HBRUSH)(COLOR_BTNSHADOW);
		    tabclassbutton.hCursor = LoadCursor(NULL, IDC_ARROW);
		    RegisterClassExW(&tabclassbutton);
		    hwndTabCloseBtn = CreateWindowExW(0, szTabButtonClass, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
											 0, 0, 0, 0,
											 hwndMain, NULL, hInstance, NULL);

	    	/*the child application window*/
		    memset(&childclass, 0, sizeof(WNDCLASSEXW));
		    childclass.lpszClassName = szChildWndClass;
		    childclass.cbSize = sizeof(WNDCLASSEXW);
		    childclass.style = CS_HREDRAW | CS_VREDRAW;
		    childclass.lpfnWndProc = MainWndProc;
		    childclass.hbrBackground = 0;
		    childclass.hCursor = LoadCursor(NULL, IDC_ARROW);
		    RegisterClassExW(&childclass);
	    	hwndChild = CreateWindowExW(0, szChildWndClass, szChildWndClass, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS,
	    		     posx, posy, wdt, hgt,
	    		     hwndMain, NULL, hInstance, NULL);
	    	EnableScrollBar(hwndChild, SB_BOTH, ESB_ENABLE_BOTH);
	    	ShowScrollBar(hwndChild, SB_BOTH, 0);


	    	/*The status bar*/
	    	lstStatusBar = list_new(16);
	    	hwndStatus = CreateWindowExW(0, szTabStatusBarClass, NULL,
					WS_CHILD | SBARS_SIZEGRIP,
					0, 0, 0, 0,
					hwndMain, NULL, hInstance, NULL);

			/*the window for the controls*/
		    memset(&wndClassDialog, 0, sizeof(WNDCLASSEXW));
		    wndClassDialog.lpszClassName = szDialogWindowClass;
		    wndClassDialog.cbSize = sizeof(WNDCLASSEXW);
		    wndClassDialog.style = CS_HREDRAW | CS_VREDRAW;
		    wndClassDialog.lpfnWndProc = DialogWndProc;
		    wndClassDialog.hbrBackground = (HBRUSH)(COLOR_BTNSHADOW);
		    wndClassDialog.hCursor = LoadCursor(NULL, IDC_ARROW);
		    wndClassDialog.hIcon = hico;
		    wndClassDialog.hIconSm = hico;
		    wndClassDialog.cbWndExtra = DLGWINDOWEXTRA;
		    RegisterClassExW(&wndClassDialog);
	    	//DWORD le = GetLastError();

	    	/*Dialog fonts */
	    	memset(&ncm, 0, sizeof(NONCLIENTMETRICS));
	    	ncm.cbSize = sizeof(NONCLIENTMETRICS);
	    	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);
	    	//DialogSepHor = ncm.iBorderWidth;
	    	//DialogSepVer = ncm.iBorderWidth;
	    	DialogSepHor = 8;
	    	DialogSepVer = 8;

	    	hDialogFont = CreateFontIndirect((const LOGFONT*)&ncm.lfCaptionFont);

	    	cursors[0] = LoadCursor(0, IDC_ARROW);
	    	cursors[1] = LoadCursor(0, IDC_IBEAM);
	    	cursors[2] = LoadCursor(0, IDC_WAIT);

			#ifdef PLATFORM_GDI_PLUS
				GdipCreateFromHDC(GetDC(hwndChild), &gdpGraphics);
			#endif



			TEXTMETRIC tm;
			if(fwplat->defaultFontSizeHint >= 10) 				__def_font_sze = fwplat->defaultFontSizeHint;
			else if(dllGetTextMetrics(GetDC(hwndChild), &tm)) 	__def_font_sze = tm.tmAscent;

			font_use(0);
            fwplat->gui.color_maxval = col_maxval;
	    	ongui_begin(fwplat);

			#ifdef PLATFORM_GDI_PLUS
			if(gdpGraphics) GdipDeleteGraphics(gdpGraphics);
			gdpGraphics = 0;
			#endif

	    	ShowWindow(hwndMain, SW_SHOWNORMAL);
	    	UpdateWindow(hwndMain);
	    } else {
	    	error_msg("Failed to create the main window");
	    	return 2;
	    }


		while(!swWinExit){
			if(!swWinExit && GetMessageW( &msg, 0, 0, 0 )){
				TranslateMessage(&msg);
				DispatchMessageW(&msg);
			}
		}

	    do{
	    	t = (char*)list_pop(fwplat->sys.lstArgv);
	    	if(t) free(t);
	    }while(t);
		fwplat->sys.lstArgv = list_destroy(fwplat->sys.lstArgv);

		__kill_timers();
	    platform_app_destroy(fwplat);

	    for(j = 0; j < lstTabPages->cnt; j++){
	    	t = (char*)list_at(lstTabPages, j);
    		if(t) free(t);
	    };
	    lstTabPages = list_destroy(lstTabPages);
    	lstToolBarItems = list_destroy(lstToolBarItems);
    	lstStatusBar = list_destroy(lstStatusBar);

    	if(hilstToolBar) dllImageList_Destroy(hilstToolBar);
    	if(hTabFont)dllDeleteObject(hTabFont);
    	FreeLibrary(hKernel32);
    	FreeLibrary(hUser32);
    	FreeLibrary(hWinDlg);
    	FreeLibrary(hWinCtrl);
		FreeLibrary(hWinGdi);

		#ifdef PLATFORM_GDI_PLUS
		GdiplusShutdown(gdpToken);
		#endif

		#ifdef PLATFORM_DIRECT2D
		dllDX2D_exit();
    	FreeLibrary(hDx2d);
		#endif


	} else {
		error_msg("Failed to load gdi32.dll or/and Comctl32.dll or/and Comdlg32.dll");
		return 3;
	}
	return 0;
}

