/*
	This file is part of package MathCandy.

    Copyright (C) 2000, 2018 Vansteenhuyse Hubert <hubert.vansteenhuyse@freecode.be>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope thapackaget it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/
#ifndef __PLATFORM_FRAMEWORK_H__
#define __PLATFORM_FRAMEWORK_H__
#include <stdarg.h>
#include <math.h>
#include <sys/time.h>

#include "muList.h"


//PORT definition
#if defined(__WINDOWS__)
	#define PORT_OS		"Microsoft\u00AE Windows\u00AE"

	#if defined(PLATFORM_GDI_PLUS)
		#define	PORT_GUI	"GDIPLUS"
	#elif defined(PLATFORM_DIRECT2D)
		#define	PORT_GUI	"Direct2D"
	#elif defined(PLATFORM_GDI)
		#define	PORT_GUI	"GDI"
	#else
		#error "Specify the makefile GUI variable: PLATFORM_GDI, PLATFORM_GDI_PLUS or PLATFORM_DIRECT2D"
	#endif

#elif defined(__LINUX__)
	#if defined(PLATFORM_GTK)
		#define PORT_OS		"Linux"
		#define PORT_GUI	"GTK"
	#else
		#error "Specify makefile GUI variable: PLATFORM_GTK"
	#endif
#else
	#error "Specify makefile PORT variable: __LINUX__ or __WINDOWS__"
#endif

#if __x86_64__
	#define PORT_BITS "64 bits"
#else
	#define PORT_BITS "32 bits"
#endif

//COMPILER definition
#if defined(__MINGW32__) || defined(__MINGW64__)
	#define COMPILER "MinGW-W64 Project"
#elif defined(__GNUC__)
	#define COMPILER "GNU Compiler Collection GCC"
#else
	#define COMPILER ""
#endif

//BUILD time definition
#define BUILD_DATE	__DATE__ " " __TIME__

//General macros and definitions
#define DEFAULT_WINDOW_SIZE 0.75f

#ifndef MAX
	#define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
	#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef ABS
	#define ABS(a) ((a) < 0 ? -(a) : (a))
#endif

#ifndef CLAMP
	#define CLAMP(val, min, max) ((val) < (min) ? (min) : (val) > (max) ? (max) : (val))
#endif


#define ARAD (1.f / 180.f * M_PI)
#define ADEG (1.f / M_PI * 180.f)
#define D2R(d) ( (d) * ARAD )			/*converts degrees to radians */
#define R2D(d) ( (d) * ADEG )			/*converts radians to degrees */

#define NOT_ZERO(i) ((int)i == 0 ? 1 : i)

//clamps d between 0 and +/- amax
#define ANGLE_CLAMP(d, amax)		( (d) - int( (int(d) / NOT_ZERO(amax)) * amax)  )
#define DEGREE_NORMALIZE(d)				( ANGLE_CLAMP(d, 360)  )
#define RADIAN_NORMALIZE(r)				( NORM_DEGREE(r * DEGREE) * RADIAN )


#define TXTOUT_MAX (1024 * 4)

typedef enum IPLAT{
	IPLAT_NULL = 0,

	/*Safe colors provided by the system*/
	SYSCOL_TEXT_F, SYSCOL_TEXT_B,
	SYSCOL_TEXT_SELETED_F, SYSCOL_TEXT_SELETED_B,
	SYSCOL_TOOLTIP_F, SYSCOL_TOOLTIP_B,
	SYSCOL_BUTTON_F, SYSCOL_BUTTON_B,

	SYSCOL_BLACK, 	SYSCOL_WHITE,
	SYSCOL_GRAY, 	SYSCOL_GRAYL,
	SYSCOL_RED, 	SYSCOL_REDL,
	SYSCOL_GREEN, 	SYSCOL_GREENL,
	SYSCOL_BLUE, 	SYSCOL_BLUEL,
	SYSCOL_BROWN, 	SYSCOL_YELLOW,
	SYSCOL_MAGENTA,	SYSCOL_MAGENTAL,
	SYSCOL_CYAN,	SYSCOL_CYANL,

	/*Redraw reason*/
	REDRAW_RESIZE,
	REDRAW_MENU_ACTION,

	/*Scroll bars*/
	SBARH,
	SBARV,

	/*Mouse*/
	MOUSE_CURSOR_DEFAULT,
	MOUSE_CURSOR_IBEAM,
	MOUSE_CURSOR_WAIT,
	MBTN_LEFT_IN,
	MBTN_LEFT_UP,
	MBTN_RIGHT_IN,
	MBTN_RIGHT_UP,
	MWHEEL_UP,
	MWHEEL_DN,

	/*Caret or cursor for the keyboard*/
	CARET_ON,
	CARET_OFF,
	CARET_NORMAL,
	CARET_OVW,

	/*Text decoration*/
	TEXT_TRANSPARENT,
	TEXT_OPAQUE,
	TEXT_BOLD,
	TEXT_ITALIC,
	TEXT_UNDERLINED,
	TEXT_DIMMED,
	TEXT_NORMAL,

	/*Toolbars*/
	TOOLBAR_ITEM_EXIT,
	TOOLBAR_ITEM_CLOSE_FILE,
	TOOLBAR_ITEM_OPENFILE,
	TOOLBAR_ITEM_DOCSAVEAS,
	TOOLBAR_ITEM_DOCSAVE,
	TOOLBAR_ITEM_DOCNEW,
	TOOLBAR_ITEM_DOCSEARCH,
	TOOLBAR_ITEM_DOCREPLACE,
	TOOLBAR_ITEM_DOCPRINT,
	TOOLBAR_ITEM_FONTSELECTION,
	TOOLBAR_ITEM_SEPARATOR,
	TOOLBAR_BUTTON,
	TOOLBAR_PUSHBUTTON,

	/*response of system dialog boxes*/
	DLG_OK,
	DLG_CANCEL,

	/*	Controls for Custom Dialog Boxes.
	 * **********************************
	 * 	The initial string (label text or desired height and width) is passed in function dlg_add_item in parameter initstring.
	 * 	The values for controls are passed with dlg_item_set_value and retrieved with dlg_item_get_value.
	 * 	Each control must be provided with an identification.
	 * 	This should be a unique value for the controls you will access later in your code.
	 */
	CTL_NEWLINE,			/*positions the next control at the begin of the next line in the dialog	value: none														returned value: none				*/
	CTL_NEWLINE_DRAW,		/*like CTL_NEWLINE, draws a line and title. initial string a label title.	value: none														returned value: none				*/
	CTL_LABEL,				/*a label. initial string: the label text									value: label text												returned value: label text			*/
	CTL_BUTTON,				/*a button. initial string: the label text.									value: none 													returned value: none				*/
	CTL_PUSH_BUTTON,		/*a push button. initial string: the label text. 							value: 1 or 0													returned value: 1 or 0				*/
	CTL_CHECK_BUTTON,		/*a check box. initial string: the label text.								value: 1 or 0													returned value: 1 or 0				*/
	CTL_TEXT,				/*a single text box. initial string: "width;" measured in characters.		value: a string													returned value: a string			*/
	CTL_TEXT_MULTI_LINE,	/*multiple lines: initial string "width;height;" measured in characters. 	value: multiple strings\n 										returned value: strings\n			*/
	CTL_LIST_BOX,			/*list box. initial string "column title;column title;..."       . 			value: multiple id;string\nstring\n...id;string\nstring\n... 	returned value: id or empty string	*/
	CTL_DROPBOX,			/*drop down box. initial string "width;" measured in characters. 			value: multiple strings\n 										returned value: string				*/

	CTL_STAT_CLICKED,		/*Status of a control, value of ctl_stat in the dialog routine.*/
	CTL_STAT_TEXT_ALTERED,

	CTL_CB_CONTINUE,		/*Possible custom dialog routine return values*/
	CTL_CB_CLOSEDLG,

	/*Drag And Drop*/
	DND_UNDEFINED,
	DND_URI_FILE,
	DND_URI_WWW,
	DND_URI_FTP,
	DND_UNC,

	/*Use Font*/
	FONT_MONOSPACE,
	FONT_SYSTEM,
	FONT_SELECTED,

}IPLAT;

typedef enum PLATKEY{
	PKNONE=0,
	PKMOUSE,
	PKAN, /*alphanumeric key*/
	PKUP, PKDN, PKLE, PKRI, PKPU, PKPD,
	PKHM, PKED, PKIN,	/*home, end, insert*/
	PKDL, PKBK,			/*delete, backspace*/
	PKPL, PKPR,			/*page left, page right*/
	PESC, PTAB, PBTAB,
	PSHIFT, PCONTROL, PALT,
	PSUPER,	/*microsoft logo key*/
	SCROLL_HALF_UP, SCROLL_HALF_DN,
}PLATKEY;

typedef struct C_RGBA{
	int r,g,b,a;
}C_RGBA;
void plat_normal_color(IPLAT icol, double *r, double *g, double *b, double *a);

typedef struct PLAT_RECT{
	int x, y, xx, yy;
}PLAT_RECT;

#define PLAT_RECT_RESET(r) ( r.x = r.y = r.xx = r.yy = 0 )
//sets based on x, y, xx and yy
#define PLAT_RECT_SET(r, ix, iy, ixx, iyy) { r.x = (ix); r.y = (iy); r.xx = (ixx); r.yy = (iyy); }
//sets based on x, y, width and height
#define PLAT_RECT_XYWH(r, ix, iy, iw, ih) { r.x = (ix); r.y = (iy); r.xx = (ix) + (iw); r.yy = (iy) + (ih); }

//returns width / height
#define PLAT_RECT_W(a) ( (a).xx - (a).x )
#define PLAT_RECT_WP(a) ( (a)->xx - (a)->x )
#define PLAT_RECT_H(a) ( (a).yy - (a).y )
#define PLAT_RECT_HP(a) ( (a)->yy - (a)->y )

//point in rectangle
#define PLAT_RECT_POINT_IN(r, ix, iy) ( (ix) > (r).x && (ix) < (r).xx && (iy) > (r).y && (iy) < (r).yy )
#define PLAT_RECT_POINT_INP(r, ix, iy) ( (ix) > (r)->x && (ix) < (r)->xx && (iy) > (r)->y && (iy) < (r)->yy )

//rectangle rectangle overlap
#define PLAT_RECT_INTERSECT(a, b) ( (a).x < (b).xx && (a).xx > (b).x && (a).y < (b).yy && (a).yy > (b).y)
#define PLAT_RECT_INTERSECTP(a, b) ( (a)->x < (b)->xx && (a)->xx > (b)->x && (a)->y < (b)->yy && (a)->yy > (b)->y)

//resizing of rectangle a, rectangle b is unchanged
extern PLAT_RECT plat_rect_grow(PLAT_RECT *a, PLAT_RECT *b);


typedef struct CTL_TOOLBAR_ITEM {
	void	*internal;
	int		ctrl_key;			/*CTRL+key to activate the item*/
	void	*response;			/*some items return a result, e.g. open file dialogs, ... */
	char	*filefilter;		/*fiter for file list, format: displayname;ext;displayname;ext;... . (eg: all files;*;text files;*.txt;)*/
	IPLAT	iplat_toolbar;		/*TOOLBAR_xxx enumerations*/
	char	*tooltip;			/*customized optional tooltip*/
	char	*text;				/*button text*/
	char	*icon;				/*icon path*/
	int		sw_active;			/*status grayed out or active*/
	int		sw_show;			/*showed or hidden*/
	int		sw_pushed;			/*pushbutton pushed or released*/
}CTL_TOOLBAR_ITEM;

/*
 * Custom Dialog Box data types.
 */
typedef void* DIALOGBOX;		/*a custom dialog box*/
typedef void* DIALOGCONTROL;	/*an item from the custom dialog box*/
typedef IPLAT (*dialog_routine)(DIALOGBOX, DIALOGCONTROL, IPLAT ctl_stat);

/**
 * Displays a single line text in a box.
 */
typedef struct CTL_INFOBOX {
	void		*internal, *internaltext;
	int			swVisible, swInvertColors;
} CTL_INFOBOX;

typedef struct CTL_STATUSBAR_ITEM {
	void	*internal;
	int		sw_active;
	char	text[1024*2];
}CTL_STATUSBAR_ITEM;

typedef struct PLATFORM_GUI{
	int	color_maxval;	/*max rgb value: 255, 65535, 16, 8, 4, 1*/
	void *sysdata;

	void (*title)(char *);
	void (*icon)(char* path);		//program icon

	void (*clrscr)(void);			//clear screen
	int (*redraw)(PLAT_RECT *r);	//if r == 0 then the entire area is redrawn

	/*Color, values between 0.0 and 1.0 */
	void (*set_intensity)(double ni);
	void (*set_fcolor)(double nr, double ng, double nb);
	void (*set_bcolor)(double nr, double ng, double nb);
	void (*set_alpha)(double na);
	void (*set_sys_color)(IPLAT fg, IPLAT bg);
	void (*swap_colors)(void);	//swaps foreground and background

	/*Caret*/
	void (*caret_set)(IPLAT caret_onoff, IPLAT caret_mode, int x, int y, char letter);

	/*Shapes*/
	void (*rectangle)(int swFilled, int x, int y, int wdt, int hgt);
	void (*circle)(int swFilled, int x, int y, int r);
	//polygons are expected to be closed shapes that can optionally be filled with background color
	void (*polygon)(int swFilled, int elements, double *x, double *y);
	//line paths are just connected lines to form an open/closed shape which cannot be filled
	void (*linepath)(int elements, double *x, double *y);
	void (*putpixel)(int x, int y);
	void (*line)(int x1, int y1, int x2, int y2);
	void (*dashline)(int x1, int y1, int x2, int y2);

	/*Text*/
	void (*textmode)(IPLAT text_mode);	/*sets transparent or opaque text draw. TEXT_xxx enumerations*/
	void (*textout)(int x, int y, char *fmt, ...);
	int (*lenpixels)(int *wdt, int *hgt, char *fmt, ...);

	/*dimensions*/
	int (*maxx)(void);
	int (*maxy)(void);
	int (*resolutionx)(void);
	int (*resolutiony)(void);

	/*fonts*/
	void (*font_use)(IPLAT ifont);
	void (*font_face)(const char *font_name, int font_size);
	int (*fontwdt)(void);
	int (*fonthgt)(void);

	/*screen shots*/
	void* (*screencopy)(int x, int y, int wdt, int hgt);
	void (*screenput)(void*, int scrx, int scry, int imgx, int imgy, int wdt, int hgt);
	void* (*screendelete)(void*);
}PLATFORM_GUI;

typedef struct PLATFORM_CONTROLS {
	void (*set_mouse_cursor)(IPLAT mouse_cursor);
	IPLAT (*get_mouse_cursor)(void);

	int (*toolbar_add)(CTL_TOOLBAR_ITEM*);			/*adds tool bar items to the tool bar, order is as the items are added*/
	void (*toolbar_refresh)(void);					/*refreshes the toolbar*/

	void (*sbar_show)(IPLAT sbar, int sw);			/*scroll bar*/
	void (*sbar_set)(IPLAT sbar, double pct);

	void* (*tabwnd_create)(char *title);			/*mdi window tab: create a document*/
	void* (*tabwnd_set_title)(void*, char *title);
	void  (*tabwnd_set_lock)(int);					/*prevents the tab to switch pages*/

	void (*statusbar_add)(CTL_STATUSBAR_ITEM*);		/*adds tool bar items, order is as the items are added*/
	void (*statusbar_refresh)(void);				/*refreshes the task bar*/

	void (*error_msg)(const char *title, char *fmt, ...);				/*system messages*/
	IPLAT(*yesno_msg)(const char *title, char *fmt, ...);
	IPLAT(*msgbox)(const char *title, char *fmt, ...);

	/**
	 * Custom dialog box routines
	 **/
	DIALOGBOX (*dlg_add_item)(DIALOGBOX, IPLAT ctlType, unsigned int id, char *text); 	/*creates and adds controls to a custom dialog, creates a new dialog box when DIALOGBOX = 0*/
	DIALOGBOX (*dlg_destroy)(DIALOGBOX);
	IPLAT (*dlg_domodal)(DIALOGBOX, char *title, dialog_routine fct);
	void (*dlg_update)(DIALOGBOX, char *title);
	void (*dlg_set_focus)(DIALOGBOX, DIALOGCONTROL);
	void (*dlg_set_response)(DIALOGBOX, IPLAT);
	IPLAT (*dlg_item_type)(DIALOGCONTROL);												/*custom dialog control query and altering */
	void (*dlg_item_show)(DIALOGCONTROL, int sw);
	int (*dlg_item_shown)(DIALOGCONTROL);
	void (*dlg_item_enable)(DIALOGCONTROL, int sw);
	int (*dlg_item_enabled)(DIALOGCONTROL);
	void (*dlg_item_set_value)(DIALOGCONTROL, char *value);
	char* (*dlg_item_get_value)(DIALOGCONTROL);
	void (*dlg_item_reset)(DIALOGCONTROL);
	unsigned int (*dlg_item_get_id)(DIALOGCONTROL);
	int (*dlg_item_get_value_size)(DIALOGCONTROL);
	DIALOGCONTROL (*dlg_item_find)(DIALOGBOX, unsigned int id);										/*custom dialog control lookup*/
	DIALOGCONTROL (*dlg_item_next)(DIALOGBOX, DIALOGCONTROL);
	void* (*dlg_get_userdata)(DIALOGBOX);
	void* (*dlg_set_userdata)(DIALOGBOX, void*);													/*attaches user usable data to the dialog, returns previous user data*/
}PLATFORM_CONTROLS;

typedef int (*__timer_routine) (void *arg);				/*return 0 to stop the timer*/
typedef struct PLATFORM_SYS {
	LIST	*lstArgv;
	void (*doevents)(void);
	void (*exit)(int exitcode);
	double (*total_memory)(void);				//returns total mem in MB

	int (*cboard_text_put)(char *buf, int len);		//returns #bytes copied from buf into the clipboard
	int (*cboard_text_sze)(void);					//returns #bytes of the clipboard content
	char* (*cboard_text_ptr)(void);					//returns the char ptr
	char* (*cboard_text_free)(char*);

	struct timeval (*time)(void);
	float (*time_delta)(struct timeval tv);
	int (*timed_call)(int msec, __timer_routine tr, void *arg);	//returns 0 on failure, 1 on success, tr is called repeatedly until it returns 0

	char*(*get_user)(void);
	char*(*get_home)(void);

	char (*get_path_seperator)(void);
	int (*mkdir)(char *path);
	int (*ls_exists)(char *path);

}PLATFORM_SYS;


enum WINSHAPE { WS_DONTCARE, WS_SQUARE };

typedef struct PLATFORM_APP {
	int 	sw_double_buffer;
	int 	wShapeHint;
	int 	defaultFontSizeHint;
	IPLAT	redraw_reason;

	/*part to be implemented by the platform implementation*/
	PLATFORM_GUI 		gui;	/*gui calls*/
	PLATFORM_CONTROLS	ctl;	/*gui controls calls*/
	PLATFORM_SYS		sys;	/*gui system calls*/

} PLATFORM_APP;

/*part to be implemented by the platform dependent code */
extern void platform_app_init(PLATFORM_APP *plat);						/* must be called on program entry */
extern void platform_app_destroy(PLATFORM_APP *plat);		/* must be called on program exit */

extern void onkey(PLATFORM_APP *plat, int key);							/*alphanumeric key pressed*/
extern void onskey(PLATFORM_APP *plat, PLATKEY key, int sw_pressed);		/*special key pressed*/
extern void onredraw(PLATFORM_APP *plat, PLAT_RECT *r);					/*screen must be redrawn*/
extern void onmouse(PLATFORM_APP *plat, int x, int y, IPLAT b);		/*mouse movement*/

/*
 * printing
 * 'rPage' is the printable area of 1 paper page
 * 'page' is the current page to be printed
 */
extern int onprint_setup(PLATFORM_APP *plat, PLAT_RECT *rPage);				/*return the amount of pages to be printed based on rPage area*/
extern int onprint_page(PLATFORM_APP *plat, PLAT_RECT *rPage, int page);	/*return 0 to abort printing or render the content of page within the rPage boundaries and return 1 to continue the print operation*/


/*controls related callbacks*/
extern void ontoolbar(PLATFORM_APP *plat, CTL_TOOLBAR_ITEM *itm);		/*toolbar item was pressed*/
extern void onscrollbar(PLATFORM_APP *plat, IPLAT bar, double pct);	/*scrollbar changed from position*/
extern int ontabclose(PLATFORM_APP *plat, void *tab); 			/*mdi tab: request to close document, return 1 to close, 0 to continue*/
extern void ontabactive(PLATFORM_APP *plat, void *tab);			/*mdi tab: user selected a document*/
extern void ondrop(PLATFORM_APP *plat, IPLAT DND_x, char *path, char *fullpath);	/*drag and drop: drop action was performed*/


extern void ongui_begin(PLATFORM_APP *plat);						/*before the gui starts*/
extern int ongui_exit(PLATFORM_APP *plat);							/*before the gui ends, return 1 to exit, 0 to continue*/



/******************************************************************************************************
 ******************************************************************************************************
 ******************************************************************************************************
 ***                            PLATFORM INDEPENDENT COMMON ROUTINES                                ***
 ***                              (implemented in platform_common.c)                                ***
 ******************************************************************************************************
 ******************************************************************************************************
 ******************************************************************************************************/

/**
 * Common macros
 **/

#ifndef LCASE
#define LCASE(c)( c >= 'A' && c <= 'Z' ? c - 'A' + 'a' : c )  						/* lowers a single char */
#endif

#ifndef UCASE
#define UCASE(c)( c >= 'a' && c <= 'z' ? c - ' ' : c )  							/* uppers a single char */
#endif

#ifndef IS_DIGIT
#define IS_DIGIT(c)( c >= '0' && c <= '9' ? 1 : 0)									/* checks if c is a digit */
#endif

#ifndef IS_HEX
#define IS_HEX(c)( IS_DIGIT(c) || (LCASE(c) >= 'a' && LCASE(c) <= 'f') ? 1 : 0)		/* checks if c is a hex */
#endif

#ifndef IS_BLANK
#define IS_BLANK(c) ( c == 0x20 || c == 0x9 || c == 0xD || c == 0xA ? 1 : 0 )
#endif

#ifndef IS_BINARY_CHAR
#define IS_BINARY_CHAR(c) (c < ' ' || c > '}' ? 1 :	0)
#endif

/**
 * Common routines
 **/
int plat_utf8_valid(const unsigned char *str);

#endif /*__PLATFORM_FRAMEWORK_H__*/
